"""utilities for working with AMDKFD driver"""

import glob

# Heap types in memory properties
#
# Imported from /usr/include/linux/kfd_sysfs.h
HEAP_TYPE_SYSTEM = 0
HEAP_TYPE_FB_PUBLIC = 1
HEAP_TYPE_FB_PRIVATE = 2
HEAP_TYPE_GPU_GDS = 3
HEAP_TYPE_GPU_LDS = 4
HEAP_TYPE_GPU_SCRATCH = 5


def parse_props(path):
    """Returns a dict corresponding to a KFD properties file"""
    with open(path) as file:
        return {key: int(value) for key, _, value in (line.partition(' ') for line in file)}


def gpus():
    """Yields GPU nodes within KFD topology and their properties"""
    for np in sorted(glob.glob('/sys/devices/virtual/kfd/kfd/topology/nodes/*')):
        props = parse_props(np + '/properties')

        # Skip CPUs
        if props['gfx_target_version'] == 0:
            continue

        yield np, props



import json
import os
import sys
from dataclasses import dataclass, field, fields
from pathlib import Path
from typing import Any, Literal, Mapping, TypeAlias

from ramalama.cli_arg_normalization import normalize_pull_arg
from ramalama.common import apple_vm, available
from ramalama.layered_config import LayeredMixin
from ramalama.log_levels import LogLevel, coerce_log_level
from ramalama.toml_parser import TOMLParser

PathStr: TypeAlias = str
DEFAULT_IMAGE: str = "quay.io/ramalama/ramalama"
DEFAULT_STACK_IMAGE: str = "quay.io/ramalama/llama-stack"
DEFAULT_RAG_IMAGE: str = "quay.io/ramalama/ramalama-rag"
SUPPORTED_ENGINES: TypeAlias = Literal["podman", "docker"]
SUPPORTED_RUNTIMES: TypeAlias = Literal["llama.cpp", "vllm", "mlx"]
COLOR_OPTIONS: TypeAlias = Literal["auto", "always", "never"]
GGUF_QUANTIZATION_MODES: TypeAlias = Literal[
    "Q2_K",
    "Q3_K_S",
    "Q3_K_M",
    "Q3_K_L",
    "Q4_0",
    "Q4_K_S",
    "Q4_K_M",
    "Q5_0",
    "Q5_K_S",
    "Q5_K_M",
    "Q6_K",
    "Q8_0",
]
DEFAULT_GGUF_QUANTIZATION_MODE: GGUF_QUANTIZATION_MODES = "Q4_K_M"


def _get_default_config_dirs() -> list[Path]:
    """Get platform-appropriate config directories."""
    dirs = [
        Path(f"{sys.prefix}/share/ramalama"),
        Path(f"{sys.prefix}/local/share/ramalama"),
    ]

    if os.name == 'nt':
        # Windows-specific paths using APPDATA and LOCALAPPDATA
        appdata = os.getenv("APPDATA", os.path.expanduser("~/AppData/Roaming"))
        localappdata = os.getenv("LOCALAPPDATA", os.path.expanduser("~/AppData/Local"))
        dirs.extend(
            [
                Path(os.path.join(localappdata, "ramalama")),
                Path(os.path.join(appdata, "ramalama")),
            ]
        )
    else:
        # Unix-specific paths
        dirs.extend(
            [
                Path("/etc/ramalama"),
                Path(os.path.expanduser(os.path.join(os.getenv("XDG_DATA_HOME", "~/.local/share"), "ramalama"))),
                Path(os.path.expanduser(os.path.join(os.getenv("XDG_CONFIG_HOME", "~/.config"), "ramalama"))),
            ]
        )

    return dirs


DEFAULT_CONFIG_DIRS = _get_default_config_dirs()


def get_default_engine() -> SUPPORTED_ENGINES | None:
    """Determine the container manager to use based on environment and platform."""
    if os.path.exists("/run/.toolboxenv"):
        return None

    if available("podman"):
        return "podman"

    return "docker" if available("docker") else None


def get_default_store() -> str:
    # Check if running as root (Unix only)
    if hasattr(os, 'geteuid') and os.geteuid() == 0:
        return "/var/lib/ramalama"

    return os.path.expanduser("~/.local/share/ramalama")


def get_all_inference_spec_dirs(subdir: str) -> list[Path]:
    ramalama_root = Path(__file__).parent.parent
    development_spec_dir = ramalama_root / "inference-spec" / subdir
    all_dirs = [development_spec_dir, *[conf_dir / "inference" for conf_dir in DEFAULT_CONFIG_DIRS]]

    return [d for d in all_dirs if d.exists()]


def get_inference_spec_files() -> dict[str, Path]:
    files: dict[str, Path] = {}

    for spec_dir in get_all_inference_spec_dirs("engines"):

        # Give preference to .yaml, then .json spec files
        file_extensions = ["*.yaml", "*.yml", "*.json"]
        for file_extension in file_extensions:
            # On naming collisions, i.e. muliple specs for one inference engine, prefer the
            # spec files discovered later (i.e. user-level > system-level)
            for spec_file in sorted(Path(spec_dir).glob(file_extension)):
                file = Path(spec_file)
                runtime = file.stem
                files[runtime] = file

    return files


def get_inference_schema_files() -> dict[str, Path]:
    files: dict[str, Path] = {}

    for schema_dir in get_all_inference_spec_dirs("schema"):

        for spec_file in sorted(Path(schema_dir).glob("schema.*.json")):
            file = Path(spec_file)
            version = file.name.replace("schema.", "").replace(".json", "")
            files[version] = file

    return files


def coerce_to_bool(value: Any) -> bool:
    if isinstance(value, bool):
        return value
    elif isinstance(value, str):
        val = value.strip().lower()
        if val in {"on", "true", "1", "yes", "y"}:
            return True
        elif val in {"off", "false", "0", "no", "n"}:
            return False
    raise ValueError(f"Cannot coerce {value!r} to bool")


@dataclass
class UserConfig:
    no_missing_gpu_prompt: bool = False

    def __post_init__(self):
        self.no_missing_gpu_prompt = coerce_to_bool(self.no_missing_gpu_prompt)


@dataclass
class RamalamaSettings:
    """These settings are not managed directly by the user"""

    config_files: list[str] | None = None


@dataclass
class RamalamaImageConfig:
    def get(self, key: str, default: Any = None) -> Any:
        return getattr(self, key, default)

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)

    def __setitem__(self, key: str, value: Any):
        setattr(self, key, value)

    def __contains__(self, key: str) -> bool:
        return key in {f.name for f in fields(self)}

    def __iter__(self):
        return iter(f.name for f in fields(self))

    def __len__(self) -> int:
        return len(fields(self))


@dataclass
class RamalamaImages(RamalamaImageConfig):
    ASAHI_VISIBLE_DEVICES: str = "quay.io/ramalama/asahi"
    ASCEND_VISIBLE_DEVICES: str = "quay.io/ramalama/cann"
    CUDA_VISIBLE_DEVICES: str = "quay.io/ramalama/cuda"
    GGML_VK_VISIBLE_DEVICES: str = "quay.io/ramalama/ramalama"
    HIP_VISIBLE_DEVICES: str = "quay.io/ramalama/rocm"
    INTEL_VISIBLE_DEVICES: str = "quay.io/ramalama/intel-gpu"
    MUSA_VISIBLE_DEVICES: str = "quay.io/ramalama/musa"
    VLLM_ASAHI_VISIBLE_DEVICES: str = "docker.io/vllm/vllm-openai"
    VLLM_ASCEND_VISIBLE_DEVICES: str = "docker.io/vllm/vllm-openai"
    VLLM_CUDA_VISIBLE_DEVICES: str = "docker.io/vllm/vllm-openai"
    VLLM_GGML_VK_VISIBLE_DEVICES: str = "docker.io/vllm/vllm-openai"
    VLLM_HIP_VISIBLE_DEVICES: str = "docker.io/vllm/vllm-openai"
    VLLM_INTEL_VISIBLE_DEVICES: str = "docker.io/vllm/vllm-openai"
    VLLM_MUSA_VISIBLE_DEVICES: str = "docker.io/vllm/vllm-openai"


@dataclass
class RamalamaRagImages(RamalamaImageConfig):
    CUDA_VISIBLE_DEVICES: str = "quay.io/ramalama/cuda-rag"
    HIP_VISIBLE_DEVICES: str = "quay.io/ramalama/rocm-rag"
    INTEL_VISIBLE_DEVICES: str = "quay.io/ramalama/intel-gpu-rag"


@dataclass
class HTTPClientConfig:
    max_retries: int = 5
    max_retry_delay: int = 30

    def __post_init__(self):
        self.max_retries = int(self.max_retries)
        if self.max_retries < 0:
            raise ValueError(f"http_client.max_retries must be non-negative: {self.max_retries}")
        self.max_retry_delay = int(self.max_retry_delay)
        if self.max_retry_delay < 0:
            raise ValueError(f"http_client.max_retry_delay must be non-negative: {self.max_retry_delay}")


@dataclass
class BaseConfig:
    api: str = "none"
    api_key: str | None = None
    cache_reuse: int = 256
    carimage: str = "registry.access.redhat.com/ubi10-micro:latest"
    container: bool = None  # type: ignore
    ctx_size: int = 0
    default_image: str = DEFAULT_IMAGE
    default_rag_image: str = DEFAULT_RAG_IMAGE
    dryrun: bool = False
    engine: SUPPORTED_ENGINES | None = field(default_factory=get_default_engine)
    env: list[str] = field(default_factory=list)
    host: str = "0.0.0.0"
    image: str = None  # type: ignore
    images: RamalamaImages = field(default_factory=RamalamaImages)
    rag_image: str | None = None
    rag_images: RamalamaRagImages = field(default_factory=RamalamaRagImages)
    keep_groups: bool = False
    max_tokens: int = 0
    ngl: int = -1
    ocr: bool = False
    port: str = "8080"
    prefix: str = None  # type: ignore
    pull: str = "newer"
    rag_format: Literal["qdrant", "json", "markdown", "milvus"] = "qdrant"
    runtime: SUPPORTED_RUNTIMES = "llama.cpp"
    selinux: bool = False
    settings: RamalamaSettings = field(default_factory=RamalamaSettings)
    stack_image: str = DEFAULT_STACK_IMAGE
    store: str = field(default_factory=get_default_store)
    summarize_after: int = 4
    temp: str = "0.8"
    thinking: bool = True
    threads: int = -1
    transport: str = "ollama"
    user: UserConfig = field(default_factory=UserConfig)
    verify: bool = True
    gguf_quantization_mode: GGUF_QUANTIZATION_MODES = DEFAULT_GGUF_QUANTIZATION_MODE
    http_client: HTTPClientConfig = field(default_factory=HTTPClientConfig)
    log_level: LogLevel | None = None

    def __post_init__(self):
        self.container = coerce_to_bool(self.container) if self.container is not None else self.engine is not None
        self.image = self.image if self.image is not None else self.default_image
        self.pull = normalize_pull_arg(self.pull, self.engine)
        self.log_level = coerce_log_level(self.log_level) if self.log_level is not None else self.log_level


class Config(LayeredMixin, BaseConfig):
    """
    Config class that combines multiple configuration layers to create a complete BaseConfig.
    Exposes the same attributes as BaseConfig, but allows for dynamic loading of configuration layers.
    Mixins should be inherited first.
    """

    def __post_init__(self):
        self._finalize_engine()
        super().__post_init__()

    def _finalize_engine(self: "Config"):
        """
        Finalizes engine selection, with special handling for Podman on macOS.

        If Podman is detected on macOS without a configured machine, it falls back on docker availability.
        """
        is_podman = self.engine is not None and os.path.basename(self.engine) == "podman"
        if is_podman and sys.platform == "darwin":
            run_with_podman_engine = apple_vm(self.engine, self)
            if not run_with_podman_engine and not self.is_set("engine"):
                self.engine = "docker" if available("docker") else None


def load_file_config() -> dict[str, Any]:
    parser = TOMLParser()
    config_path = os.getenv("RAMALAMA_CONFIG")

    if config_path and os.path.exists(config_path):
        config = parser.parse_file(config_path)
        config = config.get("ramalama", {})
        config['settings'] = {'config_files': [config_path]}
        if log_level := config.get("log_level"):
            config["log_level"] = coerce_log_level(log_level)
        return config

    config = {}
    default_config_paths = [os.path.join(conf_dir, "ramalama.conf") for conf_dir in DEFAULT_CONFIG_DIRS]

    config_paths = []
    for path in default_config_paths:
        if os.path.exists(path):
            config_paths.append(str(path))
            parser.parse_file(path)
        path_str = f"{path}.d"
        if os.path.isdir(path_str):
            for conf_file in sorted(Path(path_str).glob("*.conf")):
                config_paths.append(str(conf_file))
                parser.parse_file(conf_file)
    config = parser.data
    if config:
        config = config.get('ramalama', {})
        config['settings'] = {'config_files': config_paths}
        if log_level := config.get("log_level"):
            config["log_level"] = coerce_log_level(log_level)
    return config


def load_env_config(env: Mapping[str, str] | None = None) -> dict[str, Any]:
    if env is None:
        env = os.environ

    config: dict[str, Any] = {}
    for k, v in env.items():
        if not k.startswith("RAMALAMA"):
            continue

        k = k[8:].lstrip('_')
        subkeys = k.split("__")

        subconf = config
        for key in subkeys[:-1]:
            conf_key = key.lower()
            subconf.setdefault(conf_key, {})
            subconf = subconf[conf_key]

        subconf[subkeys[-1].lower()] = v

    if container := config.pop('in_container', None):
        config['container'] = coerce_to_bool(container)

    if container_engine := config.pop('container_engine', None):
        config['engine'] = container_engine

    if 'env' in config:
        config['env'] = config['env'].split(',')

    for key in ['images', 'rag_images']:
        if key in config:
            config[key] = json.loads(config[key])

    for key in ['ocr', 'keep_groups', 'container', 'verify']:
        if key in config:
            config[key] = coerce_to_bool(config[key])

    for key in ['threads', 'ctx_size', 'ngl', 'summarize_after']:
        if key in config:
            config[key] = int(config[key])
    if log_level := config.get("log_level"):
        config["log_level"] = coerce_log_level(log_level)
    return config


def default_config(env: Mapping[str, str] | None = None) -> Config:
    """Returns a default Config object with all layers initialized."""
    return Config(load_env_config(env), load_file_config())


CONFIG = default_config()
DEFAULT_PORT: int = int(CONFIG.port)
DEFAULT_PORT_RANGE: tuple[int, int] = (DEFAULT_PORT, DEFAULT_PORT + 100)



import argparse
import ast
import json
from pathlib import Path
from typing import Any

import jinja2
import jsonschema
import yaml

from ramalama.command import context, error, schema
from ramalama.config import get_inference_schema_files, get_inference_spec_files


def is_truthy(resolved_stmt: str) -> bool:
    return resolved_stmt not in ["None", "False", "", "[]", "{}"]


class CommandFactory:

    def __init__(self, spec_files: dict[str, Path], schema_files: dict[str, Path]):
        self.spec_files = spec_files
        self.schema_files = schema_files

    def create(self, runtime: str, command: str, ctx: context.RamalamaCommandContext) -> list[str]:
        spec_file = self.spec_files.get(runtime, None)
        if spec_file is None:
            raise FileNotFoundError(f"No specification file found for runtime '{runtime}' ")
        spec_data = CommandFactory.load_file(spec_file)
        if schema.VERSION_FIELD not in spec_data:
            raise error.InvalidInferenceEngineSpecError(
                str(spec_file), f"Missing required field '{schema.VERSION_FIELD}' "
            )

        schema_file = self.schema_files.get(spec_data[schema.VERSION_FIELD].replace(".", "-"), None)
        if schema_file is None:
            raise FileNotFoundError(f"No schema file found for spec version '{spec_data[schema.VERSION_FIELD]}' ")

        schema_data = CommandFactory.load_file(schema_file)

        try:
            CommandFactory.validate_spec(spec_data, schema_data)
        except Exception as ex:
            raise error.InvalidInferenceEngineSpecError(str(spec_file), str(ex)) from ex

        spec = schema.CommandSpecV1.from_dict(spec_data, command)
        if spec is None:
            raise NotImplementedError(f"The specification for '{runtime}' does not implement command '{command}' ")

        return CommandFactory.resolve_cmd(spec, ctx)

    @staticmethod
    def resolve_cmd(spec: schema.CommandSpecV1, ctx: context.RamalamaCommandContext) -> list[str]:
        engine = spec.command.engine

        cmd = [engine.binary]
        for option in engine.options:
            should_add = option.condition is None or is_truthy(CommandFactory.eval_stmt(option.condition, ctx))
            if not should_add:
                continue

            if option.value is None:
                cmd.append(option.name)
                continue

            value = CommandFactory.eval_stmt(option.value, ctx)
            if is_truthy(value):
                if option.name:
                    cmd.append(option.name)

                if value.startswith("[") and value.endswith("]"):
                    cmd.extend(str(v) for v in ast.literal_eval(value))
                else:
                    cmd.append(str(value))

        return cmd

    @staticmethod
    def eval_stmt(stmt: str, ctx: context.RamalamaCommandContext) -> Any:
        if not ("{{" in stmt and "}}" in stmt):
            return stmt

        return jinja2.Template(stmt).render(
            {
                "args": ctx.args,
                "model": ctx.model,
                "host": ctx.host,
            }
        )

    @staticmethod
    def validate_spec(spec_data: dict, schema: dict):
        jsonschema.validate(instance=spec_data, schema=schema)

    @staticmethod
    def load_file(path: Path) -> dict:
        if not path.exists():
            raise FileNotFoundError(f"File '{path}' not found")

        with open(path, "r") as f:
            if path.suffix == ".json":
                return json.load(f)
            elif path.suffix in [".yaml", ".yml"]:
                return yaml.safe_load(f)

            raise NotImplementedError(f"File extension '{path.suffix}' not supported")


def assemble_command(cli_args: argparse.Namespace) -> list[str]:
    cmd_factory = CommandFactory(get_inference_spec_files(), get_inference_schema_files())
    runtime = str(cli_args.runtime)
    command = str(cli_args.subcommand)
    ctx = context.RamalamaCommandContext.from_argparse(cli_args)
    return cmd_factory.create(runtime, command, ctx)



import copy
import os
import shutil
import subprocess
import tempfile
from textwrap import dedent

import ramalama.annotations as annotations
from ramalama.common import exec_cmd, perror, run_cmd, set_accel_env_vars
from ramalama.engine import BuildEngine, Engine, dry_run
from ramalama.oci_tools import engine_supports_manifest_attributes
from ramalama.transports.base import Transport

prefix = "oci://"

ociimage_raw = "org.containers.type=ai.image.model.raw"
ociimage_car = "org.containers.type=ai.image.model.car"


class OCI(Transport):
    type = "OCI"

    def __init__(self, model: str, model_store_path: str, conman: str, ignore_stderr: bool = False):
        super().__init__(model, model_store_path)

        if not conman:
            raise ValueError("RamaLama OCI Images requires a container engine")

        self.conman = conman
        self.ignore_stderr = ignore_stderr

    def login(self, args):
        conman_args = [self.conman, "login"]
        if str(args.tlsverify).lower() == "false":
            conman_args.extend([f"--tls-verify={args.tlsverify}"])
        if args.authfile:
            conman_args.extend([f"--authfile={args.authfile}"])
        if args.username:
            conman_args.extend([f"--username={args.username}"])
        if args.password:
            conman_args.extend([f"--password={args.password}"])
        if args.passwordstdin:
            conman_args.append("--password-stdin")
        if args.REGISTRY:
            conman_args.append(args.REGISTRY.removeprefix(prefix))
        return exec_cmd(conman_args)

    def logout(self, args):
        conman_args = [self.conman, "logout"]
        conman_args.append(self.model)
        return exec_cmd(conman_args)

    def _convert_to_gguf(self, outdir, source_model, args):
        with tempfile.TemporaryDirectory(prefix="RamaLama_convert_src_") as srcdir:
            ref_file = source_model.model_store.get_ref_file(source_model.model_tag)
            for file in ref_file.files:
                blob_file_path = source_model.model_store.get_blob_file_path(file.hash)
                shutil.copyfile(blob_file_path, os.path.join(srcdir, file.name))
            engine = Engine(args)
            engine.add_volume(srcdir, "/model")
            engine.add_volume(outdir.name, "/output", opts="rw")
            args.model = source_model
            engine.add_args(args.rag_image)
            # import here to avoid circular references
            from ramalama.command.factory import assemble_command

            engine.add_args(*assemble_command(args))
            if args.dryrun:
                engine.dryrun()
            else:
                engine.run()
        return self._quantize(source_model, args, outdir.name)

    def _quantize(self, source_model, args, model_dir):
        engine = Engine(args)
        engine.add_volume(model_dir, "/model", opts="rw")
        engine.add_args(args.image)
        # import here to avoid circular references
        from ramalama.command.factory import assemble_command

        args = copy.copy(args)
        args.subcommand = "quantize"
        engine.add_args(*assemble_command(args))
        if args.dryrun:
            engine.dryrun()
        else:
            engine.run()
        return f"{source_model.model_name}-{args.gguf}.gguf"

    def build_image(self, cfile, contextdir, args):
        if args.type == "car":
            parent = args.carimage
            label = ociimage_car
        else:
            parent = "scratch"
            label = ociimage_raw
        footer = dedent(
            f"""
            FROM {parent}
            LABEL {label}
            COPY --from=build /data/ /
            """
        ).strip()
        full_cfile = cfile + "\n\n" + footer + "\n"
        if args.debug:
            perror(f"Containerfile: \n{full_cfile}")
        engine = BuildEngine(args)
        return engine.build_containerfile(full_cfile, contextdir)

    def _gguf_containerfile(self, model_file_name, args):
        return dedent(
            f"""
            FROM {args.carimage} AS build
            COPY {model_file_name} /data/models/{model_file_name}
            RUN ln -s {model_file_name} /data/models/model.file
            """
        ).strip()

    def _generate_containerfile(self, source_model, args):
        # Generate the containerfile content
        # Keep this in sync with docs/ramalama-oci.5.md !
        content = [f"FROM {args.carimage} AS build"]

        model_name = source_model.model_name
        ref_file = source_model.model_store.get_ref_file(source_model.model_tag)

        for file in ref_file.files:
            blob_file_path = source_model.model_store.get_blob_file_path(file.hash)
            blob_file_path = os.path.relpath(blob_file_path, source_model.model_store.blobs_directory)
            content.append(f"COPY {blob_file_path} /data/models/{model_name}/{file.name}")
        name = ref_file.model_files[0].name if ref_file.model_files else model_name
        content.append(f"RUN ln -s {model_name}/{name} /data/models/model.file")

        return "\n".join(content)

    def build(self, source_model, args):
        gguf_dir = None
        if getattr(args, "gguf", None):
            perror("Converting to gguf ...")
            gguf_dir = tempfile.TemporaryDirectory(prefix="RamaLama_convert_", delete=False)
            contextdir = gguf_dir.name
            model_file_name = self._convert_to_gguf(gguf_dir, source_model, args)
            content = self._gguf_containerfile(model_file_name, args)
        else:
            # use blobs directory as context since paths in Containerfile are relative to it
            contextdir = source_model.model_store.blobs_directory
            content = self._generate_containerfile(source_model, args)
        try:
            return self.build_image(content, contextdir, args)
        finally:
            if gguf_dir:
                gguf_dir.cleanup()

    def tag(self, imageid, target, args):
        # Tag imageid with target
        cmd_args = [
            self.conman,
            "tag",
            imageid,
            target,
        ]
        if args.dryrun:
            dry_run(cmd_args)
        else:
            run_cmd(cmd_args)

    def _create_manifest_without_attributes(self, target, imageid, args):
        # Create manifest list for target with imageid
        cmd_args = [
            self.conman,
            "manifest",
            "create",
            target,
            imageid,
        ]
        if args.dryrun:
            dry_run(cmd_args)
        else:
            run_cmd(cmd_args)

    def _create_manifest(self, target, imageid, args):
        if not engine_supports_manifest_attributes(args.engine):
            return self._create_manifest_without_attributes(target, imageid, args)

        # Create manifest list for target with imageid
        cmd_args = [
            self.conman,
            "manifest",
            "create",
            target,
            imageid,
        ]
        if args.dryrun:
            dry_run(cmd_args)
        else:
            run_cmd(cmd_args)

        # Annotate manifest list
        cmd_args = [
            self.conman,
            "manifest",
            "annotate",
            "--annotation",
            f"{annotations.AnnotationModel}=true",
            "--annotation",
            ociimage_car if args.type == "car" else ociimage_raw,
            "--annotation",
            f"{annotations.AnnotationTitle}={args.SOURCE}",
            target,
            imageid,
        ]
        if args.dryrun:
            dry_run(cmd_args)
        else:
            run_cmd(cmd_args, stdout=None)

    def _convert(self, source_model, args):
        set_accel_env_vars()
        perror(
            f"Converting {source_model.model_store.model_name} ({source_model.model_store.model_type}) to "
            f"{self.model_store.model_name} ({self.model_store.model_type}) ..."
        )
        try:
            rm_cmd = [self.conman, "manifest", "rm", self.model]
            if args.dryrun:
                dry_run(rm_cmd)
            else:
                run_cmd(rm_cmd, ignore_stderr=True, stdout=None)
        except subprocess.CalledProcessError:
            pass
        perror(f"Building {self.model} ...")
        imageid = self.build(source_model, args)
        if args.dryrun:
            imageid = "a1b2c3d4e5f6"
        try:
            self._create_manifest(self.model, imageid, args)
        except subprocess.CalledProcessError as e:
            perror(
                f"""\
Failed to create manifest for OCI {self.model} : {e}
Tagging build instead
                """
            )
            self.tag(imageid, self.model, args)

    def convert(self, source_model, args):
        self._convert(source_model, args)

    def push(self, source_model, args):
        target = self.model
        source = source_model.model

        perror(f"Pushing {self.model} ...")
        conman_args = [self.conman, "push"]
        if args.authfile:
            conman_args.extend([f"--authfile={args.authfile}"])
        if str(args.tlsverify).lower() == "false":
            conman_args.extend([f"--tls-verify={args.tlsverify}"])
        conman_args.extend([target])
        if source != target:
            self._convert(source_model, args)
        try:
            run_cmd(conman_args)
        except subprocess.CalledProcessError as e:
            perror(f"Failed to push OCI {target} : {e}")
            raise e

    def pull(self, args):
        if not args.engine:
            raise NotImplementedError("OCI images require a container engine like Podman or Docker")

        conman_args = [args.engine, "pull"]
        if args.quiet:
            conman_args.extend(['--quiet'])
        else:
            # Write message to stderr
            perror(f"Downloading {self.model} ...")
        if str(args.tlsverify).lower() == "false":
            conman_args.extend([f"--tls-verify={args.tlsverify}"])
        if args.authfile:
            conman_args.extend([f"--authfile={args.authfile}"])
        conman_args.extend([self.model])
        run_cmd(conman_args, ignore_stderr=self.ignore_stderr)

    def remove(self, args, ignore_stderr=False):
        if self.conman is None:
            raise NotImplementedError("OCI Images require a container engine")

        try:
            conman_args = [self.conman, "manifest", "rm", self.model]
            run_cmd(conman_args, ignore_stderr=self.ignore_stderr)
        except subprocess.CalledProcessError:
            conman_args = [self.conman, "rmi", f"--force={args.ignore}", self.model]
            run_cmd(conman_args, ignore_stderr=self.ignore_stderr)

    def exists(self) -> bool:
        if self.conman is None:
            return False

        conman_args = [self.conman, "image", "inspect", self.model]
        try:
            run_cmd(conman_args, ignore_stderr=self.ignore_stderr)
            return True
        except Exception:
            return False



#!/usr/bin/env python3

import argparse
import signal
import socketserver
import threading
from datetime import datetime, timedelta

from ramalama.config import CONFIG
from ramalama.daemon.handler.ramalama import RamalamaHandler
from ramalama.daemon.logging import configure_logger, logger
from ramalama.daemon.service.model_runner import ModelRunner
from ramalama.log_levels import LogLevel


class ShutdownHandler:
    def __init__(self, server: "RamalamaServer") -> None:
        self.server = server
        self.idle_check_timer = None

    def handle_kill(self, signum, frame):
        if self.idle_check_timer:
            self.idle_check_timer.cancel()
        self.server.shutdown()

    def handle_alarm(self, signum, frame):
        # check for expiration of all models, stopping them if necessary and
        # stop shutdown server if no models are running
        self.server.check_model_expiration()
        if not self.server.model_runner.managed_models:
            self.server.shutdown()
            return

        # register alarm again for next check (Unix only)
        if hasattr(signal, 'alarm'):
            signal.alarm(self.server.idle_check_interval.seconds)

    def __enter__(self):
        signal.signal(signal.SIGINT, self.handle_kill)
        signal.signal(signal.SIGTERM, self.handle_kill)

        # SIGALRM is Unix-only, use threading.Timer on Windows as fallback
        if hasattr(signal, 'SIGALRM'):
            # set initial idle check to 300s == 5min to prevent service from stopping
            # right after being started
            signal.signal(signal.SIGALRM, self.handle_alarm)
            signal.alarm(300)
        else:
            # On Windows, use a background thread for idle checks
            self._start_idle_check_thread()

    def _start_idle_check_thread(self, interval_seconds=300):
        """Start a background thread for idle checking on Windows (no SIGALRM support)."""

        def check_and_reschedule():
            self.server.check_model_expiration()
            if not self.server.model_runner.managed_models:
                self.server.shutdown()
                return
            # Schedule next check
            self.idle_check_timer = threading.Timer(self.server.idle_check_interval.seconds, check_and_reschedule)
            self.idle_check_timer.daemon = True
            self.idle_check_timer.start()

        self.idle_check_timer = threading.Timer(interval_seconds, check_and_reschedule)
        self.idle_check_timer.daemon = True
        self.idle_check_timer.start()

    def __exit__(self, type, value, traceback):
        if self.idle_check_timer:
            self.idle_check_timer.cancel()


class RamalamaServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    def __init__(
        self, host: str, port: int, model_store_path: str, idle_check_interval: timedelta, bind_and_activate=True
    ):
        # Do not pass a RequestHandlerClass here, we will create a custom handler in finish_request
        super().__init__((host, port), None, bind_and_activate)  # type: ignore

        self.model_store_path: str = model_store_path
        self.model_runner: ModelRunner = ModelRunner()
        self.idle_check_interval = idle_check_interval

        self.allow_reuse_address = True

    def finish_request(self, request, client_address):
        RamalamaHandler(self.model_store_path, self.model_runner, request, client_address, self)

    def check_model_expiration(self):
        curr_time = datetime.now()
        for name, m in self.model_runner.managed_models.items():
            if m.expiration_date > curr_time:
                continue

            try:
                logger.info(f"Stopping expired model '{name}'...")
                self.model_runner.stop_model(m.id)
            except Exception as e:
                logger.error(f"Failed to stop expired model '{name}': {e}")

    def shutdown(self):
        logger.info("Shutting down ramalama daemon...")

        for name, managed_model in self.model_runner.managed_models.items():
            try:
                logger.info(f"Stopping model runner {name}...")
                self.model_runner.stop_model(managed_model.id)
            except Exception as e:
                logger.error(f"Error stopping model runner {name}: {e}")

        super().shutdown()


def parse_args():
    parser = argparse.ArgumentParser(description="Ramalama Daemon")
    parser.add_argument("--host", type=str, default="0.0.0.0")
    parser.add_argument("--port", type=int, default=8080)
    parser.add_argument("--model-store-path", type=str, default="/models")

    return parser.parse_args()


def run(host: str = "0.0.0.0", port: int = 8080, model_store_path: str = "/models"):
    configure_logger(CONFIG.log_level or LogLevel.DEBUG)
    logger.debug(f"Starting Ramalama daemon on {host}:{port}...")
    with RamalamaServer(host, port, model_store_path, timedelta(seconds=10)) as httpd:
        with ShutdownHandler(httpd):
            server_thread = threading.Thread(target=httpd.serve_forever, daemon=True)
            server_thread.start()
            server_thread.join()


if __name__ == '__main__':
    args = parse_args()
    run(args.host, args.port, args.model_store_path)



from dataclasses import make_dataclass
from typing import List, Protocol, get_type_hints

from ramalama.config import COLOR_OPTIONS, SUPPORTED_ENGINES, SUPPORTED_RUNTIMES, PathStr


def protocol_to_dataclass(proto_cls):
    hints = get_type_hints(proto_cls)
    fields = [(name, typ) for name, typ in hints.items()]
    return make_dataclass(f"{proto_cls.__name__}DC", fields)


class EngineArgType(Protocol):
    engine: SUPPORTED_ENGINES | None


EngineArgs = protocol_to_dataclass(EngineArgType)


class ContainerArgType(Protocol):
    container: bool | None


ContainerArgs = protocol_to_dataclass(ContainerArgType)


class StoreArgType(Protocol):
    engine: SUPPORTED_ENGINES | None
    container: bool
    store: str


StoreArgs = protocol_to_dataclass(StoreArgType)


class DefaultArgsType(Protocol):
    container: bool
    runtime: SUPPORTED_RUNTIMES
    store: PathStr
    debug: bool
    quiet: bool
    dryrun: bool
    engine: SUPPORTED_ENGINES
    noout: bool | None


DefaultArgs = protocol_to_dataclass(DefaultArgsType)


class ChatSubArgsType(Protocol):
    prefix: str
    url: str
    color: COLOR_OPTIONS
    list: bool
    model: str | None
    rag: str | None
    api_key: str | None
    ARGS: List[str] | None


ChatSubArgs = protocol_to_dataclass(ChatSubArgsType)


class ChatArgsType(DefaultArgsType, ChatSubArgsType):
    pass



import glob
import json
import os
import subprocess
import sys
import time
from abc import ABC, abstractmethod
from http.client import HTTPConnection, HTTPException
from tempfile import NamedTemporaryFile
from typing import Any, Callable

# Live reference for checking global vars
import ramalama.common
from ramalama.common import check_nvidia, exec_cmd, get_accel_env_vars, perror, run_cmd
from ramalama.logger import logger


class BaseEngine(ABC):
    """General-purpose engine for running podman or docker commands"""

    def __init__(self, args):
        base = os.path.basename(args.engine)
        self.use_docker = base == "docker"
        self.use_podman = base == "podman"
        self.args = args
        self.exec_args = [self.args.engine]
        self.base_args()
        self.add_labels()
        self.add_network()
        self.add_oci_runtime()
        self.add_privileged_options()
        self.add_pull_newer()
        self.handle_podman_specifics()

    @abstractmethod
    def base_args(self): ...

    def add_label(self, label):
        self.add(["--label", label])

    def add_name(self, name):
        self.add(["--name", name])

    def add_labels(self):
        add_labels(self.args, self.add_label)

    def add_pull(self, value: str) -> None:
        self.add_args("--pull", value)

    def add_pull_newer(self):
        if not self.args.dryrun and self.use_docker and getattr(self.args, "pull", None) == "newer":
            try:
                if not self.args.quiet:
                    perror(f"Checking for newer image {self.args.image}")
                run_cmd([str(self.args.engine), "pull", "-q", self.args.image], ignore_all=True)
            except Exception:  # Ignore errors, the run command will handle it.
                pass
        elif getattr(self.args, "pull", None):
            self.add_pull(self.args.pull)

    def add_network(self):
        if getattr(self.args, "network", None):
            self.exec_args += ["--network", self.args.network]

    def add_oci_runtime(self):
        if getattr(self.args, "oci_runtime", None):
            self.exec_args += ["--runtime", self.args.oci_runtime]
            return
        if check_nvidia() == "cuda":
            if self.use_docker:
                self.exec_args += ["--runtime", "nvidia"]
            elif os.access("/usr/bin/nvidia-container-runtime", os.X_OK):
                self.exec_args += ["--runtime", "/usr/bin/nvidia-container-runtime"]

    def add_privileged_options(self):
        if not getattr(self.args, "selinux", False):
            self.add_args("--security-opt=label=disable")
        if not getattr(self.args, "nocapdrop", False):
            self.add_args("--cap-drop=all")
            self.add_args("--security-opt=no-new-privileges")

    def add_device_options(self):
        request_no_device = getattr(self.args, "device", None) == ['none']
        if request_no_device:
            return

        if getattr(self.args, "device", None):
            for device_arg in self.args.device:
                self.exec_args += ["--device", device_arg]

        if ramalama.common.podman_machine_accel:
            self.exec_args += ["--device", "/dev/dri"]

        for path in ["/dev/dri", "/dev/kfd", "/dev/accel", "/dev/davinci*", "/dev/devmm_svm", "/dev/hisi_hdc"]:
            for dev in glob.glob(path):
                self.exec_args += ["--device", dev]

        for k, v in get_accel_env_vars().items():
            # Special case for Cuda
            if k == "CUDA_VISIBLE_DEVICES":
                if self.use_docker:
                    self.exec_args += ["--gpus", "all"]
                else:
                    # newer Podman versions support --gpus=all, but < 5.0 do not
                    self.exec_args += ["--device", "nvidia.com/gpu=all"]
            elif k == "MUSA_VISIBLE_DEVICES":
                self.exec_args += ["--env", "MTHREADS_VISIBLE_DEVICES=all"]

            self.exec_args += ["-e", f"{k}={v}"]

    def handle_podman_specifics(self):
        if getattr(self.args, "podman_keep_groups", None):
            self.exec_args += ["--group-add", "keep-groups"]

    def add(self, newargs):
        self.exec_args += newargs

    def add_args(self, *args: str) -> None:
        self.add(args)

    def add_volume(self, src: str, dest: str, *, opts="ro"):
        self.add_args("-v", f"{src}:{dest}:{opts}{self.relabel()}")

    def dryrun(self):
        dry_run(self.exec_args)

    def run(self):
        run_cmd(self.exec_args, stdout=None)

    def run_process(self) -> subprocess.CompletedProcess:
        """Run the command and return the CompletedProcess."""
        return run_cmd(self.exec_args, encoding="utf-8")

    def exec(self, stdout2null: bool = False, stderr2null: bool = False):
        exec_cmd(self.exec_args, stdout2null, stderr2null)

    def relabel(self):
        if getattr(self.args, "selinux", False) and self.use_podman:
            return ",z"
        return ""


class Engine(BaseEngine):
    """Engine for executing 'podman run'"""

    def __init__(self, args):
        super().__init__(args)
        self.add_detach_option()
        self.add_device_options()
        self.add_env_options()
        self.add_port_option()
        self.add_tty_option()

    def base_args(self) -> None:
        # For run command, do not use --rm to auto-remove container on exit
        subcommand = getattr(self.args, "subcommand", "")
        if subcommand == "run":
            self.add_args("run")
        else:
            self.add_args("run", "--rm")

    def add_name(self, name: str) -> None:
        self.add_args("--name", name)

    def add_detach_option(self) -> None:
        if getattr(self.args, "detach", False):
            self.add_args("-d")

    def add_env_option(self, value: str) -> None:
        self.add_args("--env", value)

    def add_env_options(self) -> None:
        for env in getattr(self.args, "env", []):
            self.add_env_option(env)

    def add_port_option(self) -> None:
        if getattr(self.args, "port", "") == "":
            return

        host = getattr(self.args, "host", "0.0.0.0")
        host = f"{host}:" if host != "0.0.0.0" else ""
        if self.args.port.count(":") > 0:
            self.add_args("-p", f"{host}{self.args.port}")
        else:
            self.add_args("-p", f"{host}{self.args.port}:{self.args.port}")

    def add_privileged_options(self) -> None:
        if getattr(self.args, "privileged", False):
            self.add_args("--privileged")
        else:
            super().add_privileged_options()

    def use_tty(self) -> bool:
        if not sys.stdin.isatty():
            return False
        if getattr(self.args, "ARGS", None):
            return False
        return getattr(self.args, "subcommand", "") == "run"

    def add_tty_option(self) -> None:
        if self.use_tty():
            self.add_args("-t")


class BuildEngine(BaseEngine):
    """Engine for executing 'podman build'"""

    def base_args(self) -> None:
        self.add_args("build", "-q", "--no-cache")
        if self.use_podman:
            self.add_args("--layers=false")

    def add_network(self) -> None:
        self.add_args("--network=none")

    def add_privileged_options(self) -> None:
        if self.use_podman:
            super().add_privileged_options()

    def add_pull(self, value: str) -> None:
        if self.use_docker:
            if value != "never":
                # docker build only accepts a --pull option with no value, meaning
                # "always try to pull any referenced images"
                self.add_args("--pull")
        else:
            # podman build does not accept a space-separated option (--pull foo), so pass it
            # as an equals-separated option (--pull=foo)
            self.add_args(f"--pull={value}")

    def build(self, cfile: str, context: str, /, *, tag: str | None = None) -> str:
        """
        Build an image using specified Containerfile path and context dir.
        If tag is provided, the image will be tagged.
        Return the ID of the built image.
        """
        if tag:
            self.add_args("-t", tag)
        self.add_args("-f", cfile, context)
        if self.args.dryrun:
            self.dryrun()
            return ""
        return self.run_process().stdout.strip()

    def build_containerfile(self, content: str, context: str, /, *, tag: str | None = None):
        """
        Build an image using the provided Containerfile content and context dir.
        If tag is provided, the image will be tagged.
        Return the ID of the built image.
        """
        with NamedTemporaryFile(delete_on_close=False) as tfile:
            tfile.write(content.encode("utf-8"))
            tfile.close()
            return self.build(tfile.name, context, tag=tag)


def dry_run(args):
    for arg in args:
        if not arg:
            continue
        if " " in arg:
            print('"%s"' % arg, end=" ")
        else:
            print("%s" % arg, end=" ")
    print()


def images(args):
    conman = str(args.engine) if args.engine is not None else None
    if conman == "" or conman is None:
        raise ValueError("no container manager (Podman, Docker) found")

    conman_args = [conman, "images"]
    if getattr(args, "noheading", False):
        conman_args += ["--noheading"]

    if getattr(args, "notrunc", False):
        conman_args += ["--no-trunc"]

    if args.format:
        conman_args += [f"--format={args.format}"]

    try:
        output = run_cmd(conman_args).stdout.decode("utf-8").strip()
        if output == "":
            return []
        return output.split("\n")
    except subprocess.CalledProcessError as e:
        perror("ramalama list command requires a running container engine")
        raise (e)


def containers(args):
    conman = str(args.engine) if args.engine is not None else None
    if conman == "" or conman is None:
        raise ValueError("no container manager (Podman, Docker) found")

    conman_args = [conman, "ps", "-a", "--filter", "label=ai.ramalama"]
    if getattr(args, "noheading", False):
        if conman == "docker" and not args.format:
            # implement --noheading by using --format
            conman_args += ["--format={{.ID}} {{.Image}} {{.Command}} {{.CreatedAt}} {{.Status}} {{.Ports}} {{.Names}}"]
        else:
            conman_args += ["--noheading"]

    if getattr(args, "notrunc", False):
        conman_args += ["--no-trunc"]

    if args.format:
        conman_args += [f"--format={args.format}"]

    try:
        output = run_cmd(conman_args).stdout.decode("utf-8").strip()
        if output == "":
            return []
        return output.split("\n")
    except subprocess.CalledProcessError as e:
        perror("ramalama list command requires a running container engine")
        raise (e)


def info(args):
    conman = str(args.engine) if args.engine is not None else None
    if conman == "" or conman is None:
        raise ValueError("no container manager (Podman, Docker) found")

    conman_args = [conman, "info", "--format", "json"]
    try:
        output = run_cmd(conman_args).stdout.decode("utf-8").strip()
        if output == "":
            return []
        return json.loads(output)
    except FileNotFoundError as e:
        return str(e)


def inspect(args, name, format=None, ignore_stderr=False):
    if not name:
        raise ValueError("must specify a container name")
    conman = str(args.engine) if args.engine is not None else None
    if conman == "" or conman is None:
        raise ValueError("no container manager (Podman, Docker) found")

    conman_args = [conman, "inspect"]
    if format:
        conman_args += ["--format", format]

    conman_args += [name]
    return run_cmd(conman_args, ignore_stderr=ignore_stderr).stdout.decode("utf-8").strip()


def logs(args, name, ignore_stderr=False):
    if not name:
        raise ValueError("must specify a container name")
    conman = str(args.engine) if args.engine is not None else None
    if not conman:
        raise ValueError("no container manager (Podman, Docker) found")

    conman_args = [conman, "logs", name]
    return run_cmd(conman_args, ignore_stderr=ignore_stderr).stdout.decode("utf-8").strip()


def stop_container(args, name, remove=False):
    if not name:
        raise ValueError("must specify a container name")
    conman = str(args.engine) if args.engine is not None else None
    if conman == "" or conman is None:
        raise ValueError("no container manager (Podman, Docker) found")

    ignore_stderr = False
    pod = ""
    try:
        pod = inspect(args, name, format="{{ .Pod }}", ignore_stderr=True)
    except Exception as e1:
        logger.debug(e1)
        try:
            pod = inspect(args, f"{name}-pod-model-server", format="{{ .Pod }}", ignore_stderr=True)
        except Exception as e2:  # Ignore errors, the stop command will handle it.
            logger.debug(e2)
            pass

    if pod != "":
        conman_args = [conman, "pod", "rm", "-t=0", "--ignore", "--force", pod]
    else:
        conman_args = [conman, "stop", "-t=0"]
        if args.ignore:
            if conman == "podman":
                conman_args += ["--ignore", str(args.ignore)]
            else:
                ignore_stderr = True

        conman_args += [name]
    try:
        run_cmd(conman_args, ignore_stderr=ignore_stderr)
    except subprocess.CalledProcessError:
        if args.ignore and conman == "docker":
            return
        else:
            raise

    # Remove the container if requested and not a pod (pods are already removed above)
    if remove and pod == "":
        conman_args = [conman, "rm"]
        if args.ignore:
            if conman == "podman":
                conman_args += ["--ignore"]
            else:
                ignore_stderr = True
        conman_args += [name]
        try:
            run_cmd(conman_args, ignore_stderr=ignore_stderr)
        except subprocess.CalledProcessError:
            if args.ignore and conman == "docker":
                return
            else:
                raise


def add_labels(args, add_label):
    label_map = {
        "MODEL": "ai.ramalama.model",
        "engine": "ai.ramalama.engine",
        "runtime": "ai.ramalama.runtime",
        "port": "ai.ramalama.port",
        "subcommand": "ai.ramalama.command",
    }
    for arg, label_prefix in label_map.items():
        if value := getattr(args, arg, None):
            add_label(f"{label_prefix}={value}")


def is_healthy(args, timeout: int = 3, model_name: str | None = None):
    """Check if the response from the container indicates a healthy status."""
    conn = None
    try:
        conn = HTTPConnection("127.0.0.1", args.port, timeout=timeout)
        if args.debug:
            conn.set_debuglevel(1)
        conn.request("GET", "/models")
        resp = conn.getresponse()
        if resp.status != 200:
            logger.debug(f"Container {args.name} returned status code {resp.status}: {resp.reason}")
            return False
        content = resp.read()
        if not content:
            logger.debug(f"Container {args.name} returned an empty response")
            return False
        body = json.loads(content)
        if "models" not in body:
            logger.debug(f"Container {args.name} does not include a model list in the response")
            return False
        model_names = [m["name"] for m in body["models"]]
        if not model_name:
            # The transport and tag is not included in the model name returned by the endpoint
            model_name = args.MODEL.split("://")[-1]
            model_name = model_name.split(":")[0]
        if not any(model_name in name for name in model_names):
            logger.debug(f'Container {args.name} does not include "{model_name}" in the model list: {model_names}')
            return False
        logger.debug(f"Container {args.name} is healthy")
        return True
    finally:
        if conn:
            conn.close()


def wait_for_healthy(args, health_func: Callable[[Any], bool], timeout=20):
    """Waits for a container to become healthy by polling its endpoint."""
    logger.debug(f"Waiting for container {args.name} to become healthy (timeout: {timeout}s)...")
    start_time = time.time()

    while time.time() - start_time < timeout:
        try:
            if health_func(args):
                return
        except (ConnectionError, HTTPException, UnicodeDecodeError, json.JSONDecodeError) as e:
            logger.debug(f"Health check of container {args.name} failed, retrying... Error: {e}")
        time.sleep(1)

    raise subprocess.TimeoutExpired(
        f"health check of container {args.name}", timeout, output=logs(args, args.name, ignore_stderr=not args.debug)
    )



import json
from datetime import datetime

import ramalama.annotations as annotations
from ramalama.arg_types import EngineArgType
from ramalama.common import engine_version, run_cmd

ocilabeltype = "org.containers.type"


def engine_supports_manifest_attributes(engine):
    if not engine or engine == "" or engine == "docker":
        return False
    if engine == "podman" and engine_version(engine) < "5":
        return False
    return True


def list_manifests(args: EngineArgType):
    if args.engine == "docker":
        return []

    conman_args = [
        args.engine,
        "images",
        "--filter",
        "manifest=true",
        "--format",
        (
            '{"name":"oci://{{ .Repository }}:{{ .Tag }}","modified":"{{ .CreatedAt }}",'
            '"size":{{ .VirtualSize }}, "ID":"{{ .ID }}"},'
        ),
    ]
    output = run_cmd(conman_args).stdout.decode("utf-8").strip()
    if output == "":
        return []

    manifests = json.loads(f"[{output[:-1]}]")
    if not engine_supports_manifest_attributes(args.engine):
        return manifests

    models = []
    for manifest in manifests:
        conman_args = [
            args.engine,
            "manifest",
            "inspect",
            manifest["ID"],
        ]
        output = run_cmd(conman_args).stdout.decode("utf-8").strip()

        if output == "":
            continue
        inspect = json.loads(output)
        if 'manifests' not in inspect:
            continue
        if not inspect['manifests']:
            continue
        img = inspect['manifests'][0]
        if 'annotations' not in img:
            continue
        if annotations.AnnotationModel in img['annotations']:
            models += [
                {
                    "name": manifest["name"],
                    "modified": manifest["modified"],
                    "size": manifest["size"],
                }
            ]
    return models


def list_models(args: EngineArgType):
    conman = args.engine
    if conman is None:
        return []

    # if engine is docker, size will be retrieved from the inspect command later
    # if engine is podman use "size":{{ .VirtualSize }}
    formatLine = '{"name":"oci://{{ .Repository }}:{{ .Tag }}","modified":"{{ .CreatedAt }}"'
    if conman == "podman":
        formatLine += ',"size":{{ .VirtualSize }}},'
    else:
        formatLine += ',"id":"{{ .ID }}"},'

    conman_args = [
        conman,
        "images",
        "--filter",
        f"label={ocilabeltype}",
        "--format",
        formatLine,
    ]
    output = run_cmd(conman_args, env={"TZ": "UTC"}).stdout.decode("utf-8").strip()
    if output == "":
        return []

    models = json.loads(f"[{output[:-1]}]")
    # exclude dangling images having no tag (i.e. <none>:<none>)
    models = [model for model in models if model["name"] != "oci://<none>:<none>"]

    # Grab the size from the inspect command
    if conman == "docker":
        # grab the size from the inspect command
        for model in models:
            conman_args = [conman, "image", "inspect", model["id"], "--format", "{{.Size}}"]
            output = run_cmd(conman_args).stdout.decode("utf-8").strip()
            # convert the number value from the string output
            model["size"] = int(output)
            # drop the id from the model
            del model["id"]

    models += list_manifests(args)
    for model in models:
        # Convert to ISO 8601 format
        parsed_date = datetime.fromisoformat(
            model["modified"].replace(" UTC", "").replace("+0000", "+00:00").replace(" ", "T")
        )
        model["modified"] = parsed_date.isoformat()

    return models



import json
import os

from ramalama.common import available, run_cmd
from ramalama.hf_style_repo_base import (
    HFStyleRepoFile,
    HFStyleRepoModel,
    HFStyleRepository,
    fetch_checksum_from_api_base,
)
from ramalama.model_store.snapshot_file import SnapshotFileType
from ramalama.path_utils import create_file_link

missing_modelscope = """This operation requires modelscope which is not available.

This tool can be installed via PyPI tools like uv, pip, pip3 or pipx. Example:

pip install modelscope
"""


def is_modelscope_available():
    """Check if modelscope is available on the system."""
    return available("modelscope")


def extract_modelscope_checksum(data):
    """Extract SHA-256 checksum from ModelScope API response."""
    parsed_data = json.loads(data)
    if sha256_checksum := parsed_data.get("Data", {}).get("MetaContent", {}).get("Sha256"):
        return sha256_checksum
    else:
        raise ValueError("SHA-256 checksum not found in the API response.")


def fetch_checksum_from_api(organization, file):
    """Fetch the SHA-256 checksum from the model's metadata API for a given file."""
    checksum_api_url = (
        f"{ModelScopeRepository.REGISTRY_URL}/api/v1/models/{organization}/repo/raw"
        f"?Revision=master&FilePath={file}&Needmeta=true"
    )

    return fetch_checksum_from_api_base(checksum_api_url, None, extract_modelscope_checksum)


class ModelScopeRepository(HFStyleRepository):

    REGISTRY_URL = "https://modelscope.cn"

    def fetch_metadata(self):
        self.blob_url = f"{ModelScopeRepository.REGISTRY_URL}/{self.organization}/resolve/master"
        self.model_hash = f"sha256:{fetch_checksum_from_api(self.organization, self.name)}"
        self.model_filename = self.name


class ModelScope(HFStyleRepoModel):

    REGISTRY_URL = "https://modelscope.cn/"
    ACCEPT = "Accept: application/vnd.docker.distribution.manifest.v2+json"

    def __init__(self, model, model_store_path):
        super().__init__(model, model_store_path)

        self.type = "modelscope"
        self.ms_available = is_modelscope_available()

    def get_cli_command(self):
        return "modelscope"

    def get_missing_message(self):
        return missing_modelscope

    def get_registry_url(self):
        return self.REGISTRY_URL

    def get_accept_header(self):
        return self.ACCEPT

    def get_repo_type(self):
        return "modelscope"

    def fetch_checksum_from_api(self, organization, file):
        return fetch_checksum_from_api(organization, file)

    def create_repository(self, name, organization, tag):
        return ModelScopeRepository(name, organization, tag)

    def get_cli_download_args(self, directory_path, model):
        return ["modelscope", "download", "--local_dir", directory_path, model]

    def _fetch_cache_path(self, cache_dir, namespace, repo):
        def normalize_repo_name(repo):
            return repo.replace(".", "___")

        return os.path.join(cache_dir, 'models', namespace, normalize_repo_name(repo))

    def in_existing_cache(self, args, target_path, sha256_checksum):
        if not self.ms_available:
            return False

        default_ms_caches = [os.path.join(os.environ['HOME'], '.cache/modelscope/hub')]
        namespace, repo = os.path.split(str(self.directory))

        for cache_dir in default_ms_caches:
            cache_path = self._fetch_cache_path(cache_dir, namespace, repo)
            if not cache_path or not os.path.exists(cache_path):
                continue

            file_path = os.path.join(cache_path, self.filename)
            if not os.path.exists(file_path):
                continue

            # Use cross-platform file linking (hardlink/symlink/copy)
            create_file_link(file_path, target_path)
            return True
        return False

    def push(self, _, args):
        if not self.ms_available:
            raise NotImplementedError(self.get_missing_message())
        proc = run_cmd(
            [
                "modelscope",
                "upload",
                "--repo-type",
                "model",
                self.directory,
                self.filename,
                "--cache_dir",
                os.path.join(args.store, "repos", "modelscope", ".cache"),
                "--local_dir",
                os.path.join(args.store, "repos", "modelscope", self.directory),
            ],
        )
        return proc.stdout.decode("utf-8")

    def _collect_cli_files(self, tempdir: str) -> tuple[str, list[HFStyleRepoFile]]:
        cache_dir = os.path.join(tempdir, ".cache", "modelscope", "download")
        files: list[HFStyleRepoFile] = []
        snapshot_hash = ""
        for entry in os.listdir(tempdir):
            entry_path = os.path.join(tempdir, entry)
            if os.path.isdir(entry_path) or entry == ".gitattributes":
                continue
            sha256 = ""
            metadata_path = os.path.join(cache_dir, f"{entry}.metadata")
            if not os.path.exists(metadata_path):
                continue
            with open(metadata_path) as metafile:
                lines = metafile.readlines()
                if len(lines) < 2:
                    continue
                sha256 = f"sha256:{lines[1].strip()}"
            if sha256 == "sha256:":
                continue
            if entry.lower() == "readme.md":
                snapshot_hash = sha256
                continue

            ms_file = HFStyleRepoFile(
                url=entry_path,
                header={},
                hash=sha256,
                type=SnapshotFileType.Other,
                name=entry,
            )
            # try to identify the model file in the pulled repo
            if entry.endswith(".gguf"):
                ms_file.type = SnapshotFileType.GGUFModel
            elif entry.endswith(".safetensors"):
                ms_file.type = SnapshotFileType.SafetensorModel
            files.append(ms_file)

        return snapshot_hash, files



from typing import Optional

VERSION_FIELD = "schema_version"


class CommandSpecV1:

    class Option:
        name: str
        description: Optional[str]
        value: Optional[str]
        required: bool
        condition: Optional[str]

        @staticmethod
        def from_dict(d: dict) -> "CommandSpecV1.Option":
            option = CommandSpecV1.Option()
            option.name = d["name"]
            option.description = d.get("description", None)
            option.value = d.get("value", None)
            option.required = d.get("required", True)
            option.condition = d.get("if", None)

            return option

    class Engine:
        name: str
        binary: str
        options: list["CommandSpecV1.Option"]

        @staticmethod
        def from_dict(d: dict) -> "CommandSpecV1.Engine":
            engine = CommandSpecV1.Engine()

            engine.name = d["name"]
            engine.binary = d["binary"]
            engine.options = []
            for option in d["options"]:
                opt = CommandSpecV1.Option.from_dict(option)
                engine.options.append(opt)

            return engine

    class Command:
        name: str
        engine: "CommandSpecV1.Engine"

        @staticmethod
        def from_dict(d: dict) -> "CommandSpecV1.Command":
            command = CommandSpecV1.Command()
            command.name = d["name"]
            command.engine = CommandSpecV1.Engine.from_dict(d["inference_engine"])
            return command

    def __init__(self, command: "CommandSpecV1.Command"):
        self.command = command

    @staticmethod
    def from_dict(d: dict, command: str) -> Optional["CommandSpecV1"]:
        for cmd in d.get("commands", []):
            if cmd["name"] == command:
                return CommandSpecV1(CommandSpecV1.Command.from_dict(cmd))

        return None



import configparser
import os
import sysconfig


class Shortnames:
    """Shortnames utility class"""

    shortnames: dict[str, str] = {}

    def __init__(self):
        data_path = sysconfig.get_path("data")
        file_paths = [
            "./shortnames/shortnames.conf",  # for development
            "./shortnames.conf",  # for development
            f"{data_path}/share/ramalama/shortnames.conf",
        ]

        if os.name == 'nt':
            # Windows-specific paths using APPDATA and LOCALAPPDATA
            appdata = os.getenv("APPDATA", os.path.expanduser("~/AppData/Roaming"))
            localappdata = os.getenv("LOCALAPPDATA", os.path.expanduser("~/AppData/Local"))
            file_paths.extend(
                [
                    os.path.join(localappdata, "ramalama", "shortnames.conf"),
                    os.path.join(appdata, "ramalama", "shortnames.conf"),
                ]
            )
        else:
            # Unix-specific paths using XDG conventions
            file_paths.extend(
                [
                    os.path.expanduser("~/.config/ramalama/shortnames.conf"),
                    os.path.expanduser("~/.local/share/ramalama/shortnames.conf"),
                    os.path.expanduser("~/.local/pipx/venvs/ramalama/share/ramalama/shortnames.conf"),
                    "/etc/ramalama/shortnames.conf",
                    "/usr/share/ramalama/shortnames.conf",
                    "/usr/local/share/ramalama/shortnames.conf",
                ]
            )

        self.paths = []
        for file_path in file_paths:
            config = configparser.ConfigParser(delimiters="=")
            config.read(file_path)
            if "shortnames" in config:
                self.paths.append(os.path.realpath(file_path))
                self.shortnames.update(config["shortnames"])

        # Remove leading and trailing quotes from keys and values
        self.shortnames = {self._strip_quotes(key): self._strip_quotes(value) for key, value in self.shortnames.items()}

    def _strip_quotes(self, s) -> str:
        return s.strip("'\"")

    def resolve(self, model) -> str | None:
        return self.shortnames.get(model, model)



from .huggingface import Huggingface, HuggingfaceRepository
from .modelscope import ModelScope, ModelScopeRepository
from .oci import OCI
from .ollama import Ollama, OllamaRepository
from .rlcr import RamalamaContainerRegistry
from .url import URL

__all__ = [
    "Huggingface",
    "HuggingfaceRepository",
    "ModelScope",
    "ModelScopeRepository",
    "OCI",
    "Ollama",
    "OllamaRepository",
    "RamalamaContainerRegistry",
    "URL",
]



"""Utilities for cross-platform path handling, especially for Windows Docker/Podman support."""

import os
import sys
from pathlib import Path, PureWindowsPath


def normalize_host_path_for_container(host_path: str) -> str:
    """
    Convert a host filesystem path to a format suitable for container volume mounts.

    On Windows with Docker Desktop:
    - Converts Windows paths (C:\\Users\\...) to Unix-style paths (/c/Users/...)
    - Handles both absolute and relative paths
    - Preserves forward slashes

    On Linux/macOS:
    - Returns the path unchanged (already in correct format)

    Args:
        host_path: The path on the host filesystem

    Returns:
        Path string suitable for use in container volume mounts

    Examples:
        Windows: "C:\\Users\\John\\models" -> "/c/Users/John/models"
        Linux: "/home/john/models" -> "/home/john/models"
    """
    if sys.platform != 'win32':
        # On Linux/macOS, paths are already in the correct format
        return host_path

    # On Windows, we need to convert to Unix-style paths for Docker
    # Docker Desktop for Windows expects paths in the format /c/Users/... instead of C:\Users\...

    # First, resolve symlinks and make the path absolute.
    path = Path(host_path).resolve()

    # Handle UNC paths (e.g., \\server\share). On Windows, these resolve to
    # absolute paths but don't have a drive letter.
    if not path.drive:
        return path.as_posix()

    # Handle paths with drive letters
    drive_letter = path.drive[0].lower()
    # path.as_posix() on Windows is 'C:/Users/...', so we partition on ':' and take the rest.
    return f"/{drive_letter}{path.as_posix().partition(':')[2]}"


def is_windows_absolute_path(path: str) -> bool:
    """
    Check if a path appears to be a Windows absolute path.

    Args:
        path: Path string to check

    Returns:
        True if the path looks like a Windows absolute path (e.g., C:\\, D:\\)
    """
    if sys.platform != 'win32':
        return False

    return PureWindowsPath(path).is_absolute()


def resolve_real_path(path: str) -> str:
    """
    Resolve a path to its real absolute path, handling symlinks.

    This is a cross-platform wrapper around os.path.realpath that ensures
    consistent behaviour on Windows and Unix systems.

    Args:
        path: Path to resolve

    Returns:
        Absolute path with symlinks resolved
    """
    return os.path.realpath(path)


def get_container_mount_path(host_path: str) -> str:
    """
    Get the properly formatted path for use in container mount arguments.

    This combines path resolution and normalization for container use.
    It resolves symlinks, makes the path absolute, and converts to
    container-friendly format on Windows.

    Args:
        host_path: Path on the host filesystem

    Returns:
        Path formatted for container mount commands

    Examples:
        Windows: "./models" -> "/c/Users/John/project/models"
        Linux: "./models" -> "/home/john/project/models"
    """
    # First resolve to absolute path with symlinks resolved
    real_path = resolve_real_path(host_path)

    # Then normalize for container use
    return normalize_host_path_for_container(real_path)


def create_file_link(src: str, dst: str) -> None:
    """
    Create a link from dst to src using the best available method for the platform.

    This function tries multiple linking strategies to handle cross-platform
    differences, especially for Windows where symlinks require special permissions:

    1. Try hardlink (works on Windows without admin, same disk space efficiency)
    2. Try symlink (works on Unix, and Windows with developer mode)
    3. Copy file as last resort (always works, but uses more disk space)

    Args:
        src: Source file path (must exist)
        dst: Destination link path (will be created)

    Raises:
        FileNotFoundError: If src doesn't exist
        OSError: If all linking methods fail

    Note:
        - Hardlinks only work on the same filesystem (fine for model store)
        - Hardlinks share the same inode, so deleting one doesn't affect the other
        - On Windows, hardlinks are preferred over symlinks for file operations
    """
    if not os.path.exists(src):
        raise FileNotFoundError(f"Source file does not exist: {src}")

    # Ensure destination directory exists
    os.makedirs(os.path.dirname(dst), exist_ok=True)

    # Remove existing destination if it exists
    if os.path.exists(dst) or os.path.islink(dst):
        os.unlink(dst)

    # Strategy 1: Try hardlink first (best for Windows, works without admin)
    try:
        os.link(src, dst)
        return
    except (OSError, NotImplementedError, AttributeError):
        # Hardlink failed - could be cross-filesystem, unsupported, or permission issue
        pass

    # Strategy 2: Try symlink (works on Unix, and Windows with developer mode)
    try:
        os.symlink(src, dst)
        return
    except (OSError, NotImplementedError, AttributeError):
        # Symlink failed - likely Windows without developer mode
        pass

    # Strategy 3: Last resort - copy the file
    # This uses more disk space but always works
    try:
        import shutil

        shutil.copy2(src, dst)
        return
    except Exception as e:
        raise OSError(f"Failed to create link from {src} to {dst}: all methods failed") from e



import os
from typing import Optional, Tuple

from ramalama.common import MNT_DIR, RAG_DIR, genname, get_accel_env_vars
from ramalama.file import PlainFile
from ramalama.version import version


class Kube:
    def __init__(
        self,
        model_name: str,
        model_paths: Tuple[str, str],
        chat_template_paths: Optional[Tuple[str, str]],
        mmproj_paths: Optional[Tuple[str, str]],
        args,
        exec_args,
    ):
        self.src_model_path, self.dest_model_path = model_paths
        self.src_chat_template_path, self.dest_chat_template_path = (
            chat_template_paths if chat_template_paths is not None else ("", "")
        )
        self.src_mmproj_path, self.dest_mmproj_path = mmproj_paths if mmproj_paths is not None else ("", "")
        self.src_model_path = self.src_model_path.removeprefix("oci://")

        self.ai_image = model_name
        if getattr(args, "name", None):
            self.name = args.name
        else:
            self.name = genname()

        self.args = args
        self.exec_args = exec_args
        self.image = args.image

    def _gen_volumes(self):
        mounts = """\
        volumeMounts:"""

        volumes = """
      volumes:"""

        if os.path.exists(self.src_model_path):
            m, v = self._gen_path_volume()
            mounts += m
            volumes += v
        else:
            mounts += f"""
        - mountPath: {MNT_DIR}
          subPath: /models
          name: model"""
            volumes += self._gen_oci_volume()

        if self.args.rag:
            m, v = self._gen_rag_volume()
            mounts += m
            volumes += v

        if self.src_chat_template_path and os.path.exists(self.src_chat_template_path):
            m, v = self._gen_chat_template_volume()
            mounts += m
            volumes += v

        if self.src_mmproj_path and os.path.exists(self.src_mmproj_path):
            m, v = self._gen_mmproj_volume()
            mounts += m
            volumes += v

        m, v = self._gen_devices()
        mounts += m
        volumes += v
        return mounts + volumes

    def _gen_devices(self):
        mounts = ""
        volumes = ""
        for dev in ["dri", "kfd"]:
            if os.path.exists("/dev/" + dev):
                mounts += f"""
        - mountPath: /dev/{dev}
          name: {dev}"""
                volumes += f"""
      - hostPath:
          path: /dev/{dev}
        name: {dev}"""
        return mounts, volumes

    def _gen_path_volume(self):
        mount = f"""
        - mountPath: {self.dest_model_path}
          name: model"""
        volume = f"""
      - hostPath:
          path: {self.src_model_path}
        name: model"""
        return mount, volume

    def _gen_oci_volume(self):
        return f"""
      - image:
          reference: {self.ai_image}
          pullPolicy: IfNotPresent
        name: model"""

    def _gen_rag_volume(self):
        mounts = f"""
        - mountPath: {RAG_DIR}
          name: rag"""

        volumes = f"""
      - image:
          reference: {self.args.rag}
          pullPolicy: IfNotPresent
        name: rag"""

        return mounts, volumes

    def _gen_chat_template_volume(self):
        mount = f"""
        - mountPath: {self.dest_chat_template_path}
          name: chat_template"""
        volume = f"""
      - hostPath:
          path: {self.src_chat_template_path}
        name: chat_template"""
        return mount, volume

    def _gen_mmproj_volume(self):
        mount = f"""
        - mountPath: {self.dest_mmproj_path}
          name: mmproj"""
        volume = f"""
      - hostPath:
          path: {self.src_mmproj_path}
        name: mmproj"""
        return mount, volume

    def __gen_ports(self):
        if not hasattr(self.args, "port"):
            return ""

        p = self.args.port.split(":", 2)
        ports = f"""\
        ports:
        - containerPort: {p[0]}"""
        if len(p) > 1:
            ports += f"""
          hostPort: {p[1]}"""

        return ports

    @staticmethod
    def __gen_env_vars():
        env_vars = get_accel_env_vars()

        if not env_vars:
            return ""

        env_spec = """\
        env:"""

        for k, v in env_vars.items():
            env_spec += f"""
        - name: {k}
          value: {v}"""

        return env_spec

    def generate(self) -> PlainFile:
        env_string = self.__gen_env_vars()
        port_string = self.__gen_ports()
        volume_string = self._gen_volumes()
        _version = version()

        content = f"""\
# Save the output of this file and use kubectl create -f to import
# it into Kubernetes.
#
# Created with ramalama-{_version}
apiVersion: v1
kind: Deployment
metadata:
  name: {self.name}
  labels:
    app: {self.name}
spec:
  replicas: 1
  selector:
    matchLabels:
      app: {self.name}
  template:
    metadata:
      labels:
        app: {self.name}
    spec:
      containers:
      - name: {self.name}
        image: {self.image}
        command: ["{self.exec_args[0]}"]
        args: {self.exec_args[1:]}
{env_string}
{port_string}
{volume_string}
"""

        return genfile(self.name, content)


def genfile(name, content) -> PlainFile:
    file_name = f"{name}.yaml"
    print(f"Generating Kubernetes YAML file: {file_name}")

    file = PlainFile(file_name)
    file.content = content
    return file



import json
import os
from dataclasses import dataclass
from enum import StrEnum
from pathlib import Path
from typing import Optional

from ramalama.common import generate_sha256, sanitize_filename
from ramalama.logger import logger


class RefFile:
    SEP = "---"
    MODEL_SUFFIX = "model"
    CHAT_TEMPLATE_SUFFIX = "chat"
    MMPROJ_SUFFIX = "mmproj"

    def __init__(self) -> None:
        self.hash: str = ""
        self.filenames: list[str] = []
        self.model_name: str = ""
        self.chat_template_name: str = ""
        self.mmproj_name: str = ""
        self._path: str = ""

    @property
    def path(self) -> str:
        return self._path

    @staticmethod
    def from_path(path: str) -> "RefFile":
        ref_file = RefFile()
        ref_file._path = path
        with open(path, "r") as file:
            ref_file.hash = file.readline().strip()
            filename = file.readline().strip()
            while filename != "":
                parts = filename.split(RefFile.SEP)
                if len(parts) != 2:
                    ref_file.filenames.append(filename)
                    filename = file.readline().strip()
                    continue

                ref_file.filenames.append(parts[0])
                if parts[1] == RefFile.MODEL_SUFFIX:
                    ref_file.model_name = parts[0]
                if parts[1] == RefFile.CHAT_TEMPLATE_SUFFIX:
                    ref_file.chat_template_name = parts[0]
                if parts[1] == RefFile.MMPROJ_SUFFIX:
                    ref_file.mmproj_name = parts[0]

                filename = file.readline().strip()
        return ref_file

    def remove_file(self, name: str):
        if name in self.filenames:
            self.filenames.remove(name)

            if self.chat_template_name == name:
                self.chat_template_name = ""
            if self.model_name == name:
                self.model_name = ""
            if self.mmproj_name == name:
                self.mmproj_name = ""

    def serialize(self) -> str:
        lines = [self.hash]
        for filename in self.filenames:
            line = f"{filename}{RefFile.SEP}"
            if filename == self.model_name:
                line = line + RefFile.MODEL_SUFFIX
            if filename == self.chat_template_name:
                line = line + RefFile.CHAT_TEMPLATE_SUFFIX
            if filename == self.mmproj_name:
                line = line + RefFile.MMPROJ_SUFFIX
            lines.append(line)
        return "\n".join(lines)

    def write_to_file(self):
        with open(self.path, "w") as file:
            file.write(self.serialize())
            file.flush()

    @staticmethod
    def map_to_refjsonfile(ref_file_path: str, snapshot_directory: str) -> "RefJSONFile":
        ref_file = RefFile.from_path(ref_file_path)

        ref = RefJSONFile(
            hash=ref_file.hash,
            path=f"{ref_file.path}.json",
            files=[],
        )

        def determine_type(filename: str) -> StoreFileType:
            if filename == ref_file.model_name:
                return StoreFileType.GGUF_MODEL
            if filename == ref_file.chat_template_name:
                return StoreFileType.CHAT_TEMPLATE
            if filename == ref_file.mmproj_name:
                return StoreFileType.MMPROJ
            return StoreFileType.OTHER

        def determine_blob_hash(filename: str) -> str:
            blob_path = Path(os.path.join(snapshot_directory, sanitize_filename(ref_file.hash), filename)).resolve()
            if not os.path.exists(blob_path):
                return generate_sha256(filename)
            return blob_path.stem

        for file in ref_file.filenames:
            ftype = determine_type(file)
            ref.files.append(StoreFile(determine_blob_hash(file), file, ftype))

        return ref


#
# Temporary migration routine to ensure smooth transition to new RefFile format
#
def migrate_reffile_to_refjsonfile(ref_file_path: str, snapshot_directory: str) -> Optional["RefJSONFile"]:
    # Check if a ref file in old format is present by removing the file extension
    old_ref_file_path = ref_file_path.replace(".json", "")
    if os.path.exists(old_ref_file_path):
        logger.debug(f"Migrating old ref file '{old_ref_file_path}' to new format")
        ref: RefJSONFile = RefFile.map_to_refjsonfile(old_ref_file_path, snapshot_directory)
        ref.write_to_file()
        try:
            os.remove(old_ref_file_path)
            return ref
        except Exception as ex:
            logger.debug(f"Failed to remove old ref file '{old_ref_file_path}'\n: {ex}")

    return None


class StoreFileType(StrEnum):
    GGUF_MODEL = "gguf"
    SAFETENSOR_MODEL = "safetensor"
    MMPROJ = "mmproj"
    CHAT_TEMPLATE = "chat_template"
    OTHER = "other"

    @staticmethod
    def from_str(s: str) -> "StoreFileType":
        if s == StoreFileType.GGUF_MODEL.value:
            return StoreFileType.GGUF_MODEL
        if s == StoreFileType.SAFETENSOR_MODEL.value:
            return StoreFileType.SAFETENSOR_MODEL
        if s == StoreFileType.MMPROJ.value:
            return StoreFileType.MMPROJ
        if s == StoreFileType.CHAT_TEMPLATE.value:
            return StoreFileType.CHAT_TEMPLATE
        return StoreFileType.OTHER


@dataclass
class StoreFile:
    hash: str
    name: str
    type: StoreFileType


@dataclass
class RefJSONFile:
    hash: str
    path: str
    files: list[StoreFile]

    version: str = "v1.0.1"

    def to_json(self) -> str:
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=2)

    def write_to_file(self):
        with open(self.path, "w") as file:
            file.write(self.to_json())
            file.flush()

    @property
    def model_files(self) -> list[StoreFile]:
        return [file for file in self.files if file.type == StoreFileType.GGUF_MODEL]

    @property
    def safetensor_model_files(self) -> list[StoreFile]:
        return [file for file in self.files if file.type == StoreFileType.SAFETENSOR_MODEL]

    @property
    def chat_templates(self) -> list[StoreFile]:
        return [file for file in self.files if file.type == StoreFileType.CHAT_TEMPLATE]

    @property
    def mmproj_files(self) -> list[StoreFile]:
        return [file for file in self.files if file.type == StoreFileType.MMPROJ]

    def remove_file(self, hash: str):
        for file in self.files:
            if file.hash == hash:
                self.files.remove(file)
                break

    @staticmethod
    def from_path(path: str) -> "RefJSONFile":
        with open(path, "r") as f:
            data = json.loads(f.read())

            ref_version = data["version"]
            ref_hash = data["hash"]
            ref_path = path
            ref_files = []
            for file in data["files"]:
                file_hash = file["hash"]
                file_name = file["name"]
                file_type = StoreFileType.from_str(file["type"])
                ref_files.append(StoreFile(file_hash, file_name, file_type))

            ref_file = RefJSONFile(
                version=ref_version,
                hash=ref_hash,
                path=ref_path,
                files=ref_files,
            )
            # ref file has moved
            if ref_file.path != data["path"]:
                logger.debug(f"Updating ref file path to '{ref_file.path}'")
                ref_file.write_to_file()
            return ref_file



import http.server
import urllib.parse
import urllib.request

from ramalama.daemon.handler.base import APIHandler
from ramalama.daemon.logging import logger
from ramalama.daemon.service.model_runner import ModelRunner
from ramalama.transports.transport_factory import CLASS_MODEL_TYPES


class ModelProxyHandler(APIHandler):

    PATH_PREFIX = "/model"

    def __init__(self, model_runner: ModelRunner):
        super().__init__(model_runner)

        self.model_runner = model_runner

    @staticmethod
    def build_proxy_path(model: CLASS_MODEL_TYPES) -> str:
        return f"{ModelProxyHandler.PATH_PREFIX}/{model.model_organization}/{model.model_name}"

    def handle_get(self, handler: http.server.SimpleHTTPRequestHandler, is_referred: bool = False):
        if handler.path == f"{ModelProxyHandler.PATH_PREFIX}":
            self._handle_get_running_models(handler)

            return

        self._forward_request(handler, is_referred)

    def handle_head(self, handler: http.server.SimpleHTTPRequestHandler, is_referred: bool = False):
        self._forward_request(handler, is_referred)

    def handle_post(self, handler: http.server.SimpleHTTPRequestHandler, is_referred: bool = False):
        self._forward_request(handler, is_referred)

    def handle_put(self, handler: http.server.SimpleHTTPRequestHandler, is_referred: bool = False):
        self._forward_request(handler, is_referred)

    def handle_delete(self, handler: http.server.SimpleHTTPRequestHandler, is_referred: bool = False):
        self._forward_request(handler, is_referred)

    def _forward_request(self, handler: http.server.SimpleHTTPRequestHandler, is_referred: bool = False):

        forward_path = ""
        proxy_path = ""
        if is_referred:
            logger.debug("request is referred")
            if "Referer" not in handler.headers:
                msg = "Something went wrong, no referer header found"
                logger.error(msg)
                handler.send_error(500, msg)
                return
            proxy_path = urllib.parse.urlparse(handler.headers["Referer"]).path
            forward_path = handler.path.replace("/".join(proxy_path.split("/")[:-1]), "", 1)
        else:
            logger.debug("request is not referred")
            proxy_path = handler.path
            forward_path = handler.path.replace(proxy_path, "", 1)

        model = self.model_runner.served_models.get(proxy_path, None)
        if model is None:
            msg = f"No model for path '{proxy_path}' found"
            logger.error(msg)
            handler.send_error(404, msg)
            return
        model.update_expiration_date()

        target_url = f"http://127.0.0.1:{model.port}{forward_path}"
        method = handler.command
        headers = handler.headers
        data = None

        if 'Content-Length' in headers:
            length = int(headers['Content-Length'])
            data = handler.rfile.read(length)

        logger.debug(f"Forwarding request -X {method} {target_url}\nHEADER: {headers} \nDATA: {data!r}")

        request = urllib.request.Request(target_url, data=data, headers=dict(headers), method=method)
        with urllib.request.urlopen(request) as response:
            handler.send_response(response.status)

            hop_by_hop_headers = [
                'connection',
                'keep-alive',
                'proxy-authenticate',
                'proxy-authorization',
                'te',
                'trailers',
                'transfer-encoding',
                'upgrade',
            ]
            for key, value in response.getheaders():
                if key.lower() in hop_by_hop_headers:
                    continue
                handler.send_header(key, value)

            handler.end_headers()
            for line in response:
                handler.wfile.write(line)
            handler.wfile.flush()

            logger.debug(f"Received response from -X {method} {target_url}\nRESPONSE: ")



import logging
import sys

from ramalama.log_levels import LogLevel, coerce_log_level

logger = logging.getLogger("ramalama")


def configure_logger(level: LogLevel | str | int = LogLevel.WARNING) -> None:
    global logger

    resolved_level = coerce_log_level(level)
    lvl_value = int(resolved_level)

    logger.setLevel(lvl_value)
    logger.propagate = False

    fmt = "%(asctime)s - %(levelname)s - %(message)s"
    datefmt = "%Y-%m-%d %H:%M:%S"
    formatter = logging.Formatter(fmt, datefmt)

    if not logger.handlers:
        logger.addHandler(logging.StreamHandler(sys.stderr))

    for handler in logger.handlers:
        handler.setLevel(lvl_value)
        handler.setFormatter(formatter)



#
# Copied from https://github.com/engelmi/go2jinja
#

import argparse
import re
import sys
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional


class NodeType(Enum):
    RANGE = "range"
    CONTINUE = "continue"
    BREAK = "break"
    IF = "if"
    ELIF = "else if"
    ELSE = "else"
    END = "end"
    CONTENT = ""
    STATEMENT = "stmt"
    ASSIGNMENT = "assignment"


@dataclass
class Node:
    start: int
    end: int
    content: str

    type: NodeType | None

    prev: Optional["Node"]
    next: Optional["Node"]
    parent: Optional["Node"]

    open_scope_node: Optional["Node"] = None
    end_scope_node: Optional["Node"] = None
    children: list["Node"] = field(default_factory=lambda: [])

    artificial: bool = False


class FunctionType(Enum):
    PLAIN = "plain"

    AND = "and"
    OR = "or"
    NOT = "not"

    EQUALS = "eq"
    NEQUALS = "ne"
    LESSER = "lt"
    LESSEREQUALS = "le"
    GREATER = "gt"
    GREATEREQUALS = "ge"

    LEN = "len"
    SLICE = "slice"
    INDEX = "index"

    PRINTF = "printf"


FUNCTION_MAPPING = {
    FunctionType.AND: "and",
    FunctionType.OR: "or",
    FunctionType.NOT: "not",
    FunctionType.EQUALS: "==",
    FunctionType.NEQUALS: "!=",
    FunctionType.LESSER: "<",
    FunctionType.LESSEREQUALS: "<=",
    FunctionType.GREATER: ">",
    FunctionType.GREATEREQUALS: ">=",
    FunctionType.LEN: "|length",
    FunctionType.SLICE: "",
    FunctionType.INDEX: "",
    FunctionType.PRINTF: "printf",
}


@dataclass
class FunctionNode:
    content: str
    type: FunctionType

    operands: list["FunctionNode"] = field(default_factory=lambda: [])

    brackets: bool = False

    def to_jinja(self) -> str:
        prefix = ""
        postfix = ""
        if self.brackets:
            prefix = "("
            postfix = ")"
        if self.type in [
            FunctionType.EQUALS,
            FunctionType.NEQUALS,
            FunctionType.GREATER,
            FunctionType.GREATEREQUALS,
            FunctionType.LESSER,
            FunctionType.LESSEREQUALS,
        ]:
            return (
                prefix
                + " or ".join(
                    [
                        f"{self.operands[0].to_jinja()}{FUNCTION_MAPPING[self.type]}{self.operands[i].to_jinja()}"
                        for i in range(1, len(self.operands))
                    ]
                )
                + postfix
            )
        elif self.type in [FunctionType.AND, FunctionType.OR]:
            return prefix + f" {FUNCTION_MAPPING[self.type]} ".join([op.to_jinja() for op in self.operands]) + postfix
        elif self.type == FunctionType.NOT:
            return prefix + f"{FUNCTION_MAPPING[self.type]} {self.operands[0].to_jinja()}" + postfix
        elif self.type == FunctionType.LEN:
            return prefix + f"({self.operands[0].to_jinja()}){FUNCTION_MAPPING[self.type]}" + postfix
        elif self.type == FunctionType.SLICE:
            s = ":"
            if len(self.operands) == 2:
                s = f"{self.operands[1].to_jinja()}:"
            elif len(self.operands) > 2:
                s = ":".join(self.operands[i].to_jinja() for i in range(1, len(self.operands)))
            return prefix + f"({self.operands[0].to_jinja()})[{s}]" + postfix
        elif self.type == FunctionType.INDEX:
            return prefix + f"({self.operands[0].to_jinja()})[{self.operands[1].to_jinja()}]" + postfix
        elif self.type == FunctionType.PRINTF:
            return (
                prefix
                + f"{self.operands[0].to_jinja()}.format({', '.join([op.to_jinja() for op in self.operands[1:]])})"
                + postfix
            )

        if self.content.startswith('"') and self.content.endswith('"'):
            return prefix + self.content.replace("\n", "\\n") + postfix
        return prefix + self.content + postfix


GO_SYMBOL_OPEN_BRACKETS = "{{"
GO_SYMBOL_CLOSE_BRACKETS = "}}"
JINJA_SYMBOL_OPEN_BRACKETS = "{%"
JINJA_SYMBOL_CLOSE_BRACKETS = "%}"
JINJA_SYMBOL_STMT_OPEN_BRACKETS = "{{"
JINJA_SYMBOL_STMT_CLOSE_BRACKETS = "}}"
SYMBOL_REMOVE_WHITESPACE = "-"

REGEX_VARIABLE = "(\\.[A-Za-z_][A-Za-z0-9_\\.]*)"
REGEX_LOCAL_VARIABLE = "(\\$\\.?[A-Za-z_][A-Za-z0-9_\\.]*)"
REGEX_NODE_START_BLOCK = f"{GO_SYMBOL_OPEN_BRACKETS}{SYMBOL_REMOVE_WHITESPACE}?\\s*"
REGEX_NODE_PIPELINE = "(.+\\n*)"
REGEX_NODE_END_BLOCK = f"\\s*{SYMBOL_REMOVE_WHITESPACE}?{GO_SYMBOL_CLOSE_BRACKETS}"
REGEX_NODE_IF = f"{REGEX_NODE_START_BLOCK}(if)\\s{REGEX_NODE_PIPELINE}{REGEX_NODE_END_BLOCK}"
REGEX_NODE_ELIF = f"{REGEX_NODE_START_BLOCK}(else if)\\s{REGEX_NODE_PIPELINE}{REGEX_NODE_END_BLOCK}"
REGEX_NODE_ELSE = f"{REGEX_NODE_START_BLOCK}(else){REGEX_NODE_END_BLOCK}"
REGEX_NODE_END = f"{REGEX_NODE_START_BLOCK}(end){REGEX_NODE_END_BLOCK}"
REGEX_NODE_RANGE = (
    f"{REGEX_NODE_START_BLOCK}(range)\\s+"
    f"({REGEX_LOCAL_VARIABLE}\\s*,\\s*{REGEX_LOCAL_VARIABLE}\\s*:=\\s*)"
    f"?{REGEX_VARIABLE}{REGEX_NODE_END_BLOCK}"
)
REGEX_NODE_CONTINUE = f"{REGEX_NODE_START_BLOCK}continue{REGEX_NODE_END_BLOCK}"  # noqa: E275
REGEX_NODE_BREAK = f"{REGEX_NODE_START_BLOCK}break{REGEX_NODE_END_BLOCK}"  # noqa: E275
REGEX_NODE_STMT = f"{REGEX_NODE_START_BLOCK}({REGEX_VARIABLE}|{REGEX_LOCAL_VARIABLE}){REGEX_NODE_END_BLOCK}"
REGEX_NODE_ASSIGNMENT = (
    f"{REGEX_NODE_START_BLOCK}" f"{REGEX_LOCAL_VARIABLE}\\s*:?=\\s*{REGEX_NODE_PIPELINE}" f"{REGEX_NODE_END_BLOCK}"
)
GO_KEYWORDS: Dict[NodeType, re.Pattern] = {
    NodeType.IF: re.compile(R"{}".format(REGEX_NODE_IF), re.S),
    NodeType.ELIF: re.compile(R"{}".format(REGEX_NODE_ELIF), re.S),
    NodeType.ELSE: re.compile(R"{}".format(REGEX_NODE_ELSE), re.S),
    NodeType.END: re.compile(R"{}".format(REGEX_NODE_END), re.S),
    NodeType.RANGE: re.compile(R"{}".format(REGEX_NODE_RANGE), re.S),
    NodeType.STATEMENT: re.compile(R"{}".format(REGEX_NODE_STMT), re.S),
    NodeType.ASSIGNMENT: re.compile(R"{}".format(REGEX_NODE_ASSIGNMENT), re.S),
    NodeType.CONTINUE: re.compile(R"{}".format(REGEX_NODE_CONTINUE), re.S),
    NodeType.BREAK: re.compile(R"{}".format(REGEX_NODE_BREAK), re.S),
}


def detect_node_type(stmt: str) -> Optional[NodeType]:

    # from most complex to least
    ordered_regex_list = [
        (NodeType.RANGE, GO_KEYWORDS[NodeType.RANGE]),
        (NodeType.IF, GO_KEYWORDS[NodeType.IF]),
        (NodeType.ELIF, GO_KEYWORDS[NodeType.ELIF]),
        (NodeType.ELSE, GO_KEYWORDS[NodeType.ELSE]),
        (NodeType.END, GO_KEYWORDS[NodeType.END]),
        (NodeType.CONTINUE, GO_KEYWORDS[NodeType.CONTINUE]),
        (NodeType.BREAK, GO_KEYWORDS[NodeType.BREAK]),
        (NodeType.ASSIGNMENT, GO_KEYWORDS[NodeType.ASSIGNMENT]),
        (NodeType.STATEMENT, GO_KEYWORDS[NodeType.STATEMENT]),
    ]

    for regex in ordered_regex_list:
        ntype, reg = regex
        if reg.match(stmt) is not None:
            return ntype
    return None


def parse_go_template(content: str) -> list[Node]:
    root_nodes: list[Node] = []

    prev_expr_node: Node | None = None
    current_scope_nodes: list[Node] = []
    start_pos = content.find(GO_SYMBOL_OPEN_BRACKETS)
    end_pos = 0
    while start_pos != -1:

        if end_pos == 0 and start_pos != 0:
            content_node = Node(
                end_pos,
                start_pos,
                content[end_pos:start_pos],
                NodeType.CONTENT,
                prev=None,
                next=None,
                parent=None,
                children=[],
                artificial=False,
            )
            root_nodes.append(content_node)
        elif start_pos - end_pos > 0:
            content_node = Node(
                end_pos,
                start_pos,
                content[end_pos:start_pos],
                NodeType.CONTENT,
                prev=None,
                next=None,
                parent=None,
                children=[],
                artificial=False,
            )
            current_scope_node = None if not current_scope_nodes else current_scope_nodes[-1]
            if current_scope_node is not None:
                content_node.parent = current_scope_node
                current_scope_node.children.append(content_node)
            else:
                root_nodes.append(content_node)

        end_pos = content.find(GO_SYMBOL_CLOSE_BRACKETS, start_pos) + len(GO_SYMBOL_CLOSE_BRACKETS)
        if end_pos == -1:
            raise Exception("Found opening without closing brackets")

        stmt = content[start_pos:end_pos]
        node_type = detect_node_type(stmt)

        expr_node = Node(
            start_pos,
            end_pos,
            content[start_pos:end_pos],
            node_type,
            prev=prev_expr_node,
            next=None,
            parent=None,
            children=[],
            artificial=False,
        )
        if prev_expr_node is not None:
            prev_expr_node.next = expr_node

        if expr_node.type in [NodeType.IF, NodeType.RANGE]:
            current_scope_node = None if not current_scope_nodes else current_scope_nodes[-1]
            if current_scope_node is not None:
                expr_node.parent = current_scope_node
            current_scope_nodes.append(expr_node)
        elif expr_node.type in [NodeType.ELIF, NodeType.ELSE]:
            if current_scope_nodes:
                prev = current_scope_nodes.pop()
                prev.end_scope_node = expr_node
                expr_node.open_scope_node = prev
            current_scope_node = None if not current_scope_nodes else current_scope_nodes[-1]
            if current_scope_node is not None:
                expr_node.parent = current_scope_node
            current_scope_nodes.append(expr_node)
        elif expr_node.type == NodeType.END:
            prev = current_scope_nodes.pop()
            prev.end_scope_node = expr_node
            expr_node.open_scope_node = prev
            current_scope_node = None if not current_scope_nodes else current_scope_nodes[-1]
            if current_scope_node is not None:
                expr_node.parent = current_scope_node
        else:
            current_scope_node = None if not current_scope_nodes else current_scope_nodes[-1]
            if current_scope_node is not None:
                expr_node.parent = current_scope_node

        if expr_node.parent is None:
            root_nodes.append(expr_node)
        else:
            expr_node.parent.children.append(expr_node)

        prev_expr_node = expr_node

        start_pos = content.find(GO_SYMBOL_OPEN_BRACKETS, end_pos)

    if end_pos < len(content):
        content_node = Node(
            end_pos,
            len(content) + 1,
            content[end_pos : len(content) + 1],
            NodeType.CONTENT,
            prev=None,
            next=None,
            parent=None,
            children=[],
            artificial=False,
        )
        root_nodes.append(content_node)

    return root_nodes


def is_jinja_template(content: str) -> bool:
    return re.compile(R".*{%\-?.+\-?%}", re.S).match(content) is not None


def is_go_template(content: str) -> bool:
    return re.compile(R".*{{\-?.+\-?}}", re.S).match(content) is not None and not is_jinja_template(content)


def go_to_jinja(content: str) -> str:
    if not is_go_template(content):
        return ""

    loop_vars: list[str] = []
    loop_index_vars: list[str] = []

    def transform_go_var_to_jinja(var: str, check_loop_vars: bool = True) -> str:
        var = var.lstrip("$.").lower()
        if check_loop_vars and loop_vars:
            var = f"{loop_vars[-1]}.{var}"
        return var

    def transform_go_local_var_to_jinja(var: str) -> str:
        if loop_index_vars and loop_index_vars[-1] == var:
            return "loop.index0"
        return transform_go_var_to_jinja(var, False).lower()

    def parse_pipeline(pipeline: str) -> str:

        def parse_variable(pipeline: str) -> str:
            reg = re.compile(R"{}".format(f"{REGEX_VARIABLE}"))
            m = reg.match(pipeline)
            if m is not None:
                start, end = m.span()
                if start == 0 and end == (len(pipeline)):
                    return transform_go_var_to_jinja(pipeline)

            reg = re.compile(R"{}".format(f"{REGEX_LOCAL_VARIABLE}"))
            m = reg.match(pipeline)
            if m is not None:
                start, end = m.span()
                if start == 0 and end == (len(pipeline)):
                    return transform_go_local_var_to_jinja(pipeline)

            return pipeline

        def parse_functions(pipeline: str) -> FunctionNode:
            if not pipeline.isspace():
                pipeline = pipeline.lstrip().rstrip()

            longest_match: FunctionType | None = None
            for ft in FUNCTION_MAPPING.keys():
                if pipeline.startswith(ft.value):
                    if longest_match is None or len(ft.value) > len(longest_match.value):
                        longest_match = ft

            if longest_match is not None:
                func_content = pipeline[len(longest_match.value) :].strip()
                node = FunctionNode(func_content, longest_match)

                quotes_open = False
                open_brackets = 0
                groups = []
                start, end = 0, 0
                i = 0
                prev_c = ""
                for c in func_content:
                    if c == "(" and not quotes_open:
                        if open_brackets == 0:
                            start = i
                        open_brackets += 1
                    elif c == ")" and not quotes_open:
                        open_brackets -= 1
                        if open_brackets == 0:
                            end = i
                            groups.append((func_content[start + 1 : end], True))
                    elif c == '"' and prev_c != "\\":
                        quotes_open = not quotes_open
                    elif c == " " and not quotes_open and open_brackets == 0 and prev_c != ")":
                        end = i
                        groups.append((func_content[start:end], False))
                        start = i + 1

                    prev_c = c
                    i += 1

                    if i == len(func_content):
                        rest = func_content[end:].lstrip(" )")
                        if rest != "":
                            groups.append((rest, False))

                for group, brackets in groups:
                    child_node = parse_functions(group)
                    child_node.brackets = brackets
                    node.operands.append(child_node)

                return node

            return FunctionNode(parse_variable(pipeline), FunctionType.PLAIN)

        return parse_functions(pipeline).to_jinja()

    def node_to_jinja_str(node: Node) -> str:
        if node.type == NodeType.STATEMENT:
            m = GO_KEYWORDS[NodeType.STATEMENT].match(node.content)
            if m is not None:
                content = transform_go_var_to_jinja(node.content[m.start(1) : m.end(1)])
                content = transform_go_local_var_to_jinja(content)
                content = node.content[: m.start(1)] + content + node.content[m.end(1) :]
                return content.replace(GO_SYMBOL_OPEN_BRACKETS, JINJA_SYMBOL_STMT_OPEN_BRACKETS).replace(
                    GO_SYMBOL_CLOSE_BRACKETS, JINJA_SYMBOL_STMT_CLOSE_BRACKETS
                )
        elif node.type == NodeType.IF:
            m = GO_KEYWORDS[NodeType.IF].match(node.content)
            if m is not None and len(m.groups()) == 2:
                pipeline = m.groups()[1].strip()
                to_parse = pipeline
                if to_parse[-1] == SYMBOL_REMOVE_WHITESPACE:
                    to_parse = to_parse[:-1]
                return (
                    node.content.replace(to_parse, parse_pipeline(to_parse))
                    .replace(GO_SYMBOL_OPEN_BRACKETS, JINJA_SYMBOL_OPEN_BRACKETS)
                    .replace(GO_SYMBOL_CLOSE_BRACKETS, JINJA_SYMBOL_CLOSE_BRACKETS)
                )
        elif node.type == NodeType.ELIF:
            m = GO_KEYWORDS[NodeType.ELIF].match(node.content)
            if m is not None and len(m.groups()) == 2:
                pipeline = m.groups()[1].strip()
                to_parse = pipeline
                if to_parse[-1] == SYMBOL_REMOVE_WHITESPACE:
                    to_parse = to_parse[:-1]
                return (
                    node.content.replace(to_parse, parse_pipeline(to_parse))
                    .replace(node.type.value, "elif")
                    .replace(GO_SYMBOL_OPEN_BRACKETS, JINJA_SYMBOL_OPEN_BRACKETS)
                    .replace(GO_SYMBOL_CLOSE_BRACKETS, JINJA_SYMBOL_CLOSE_BRACKETS)
                )
        elif node.type == NodeType.ELSE:
            return node.content.replace(GO_SYMBOL_OPEN_BRACKETS, JINJA_SYMBOL_OPEN_BRACKETS).replace(
                GO_SYMBOL_CLOSE_BRACKETS, JINJA_SYMBOL_CLOSE_BRACKETS
            )
        elif node.type == NodeType.END:
            m = GO_KEYWORDS[NodeType.END].match(node.content)
            if m is None:
                return ""

            if node.open_scope_node and node.open_scope_node.type in [NodeType.IF, NodeType.ELIF, NodeType.ELSE]:
                return (
                    node.content[: m.start(1)].replace(GO_SYMBOL_OPEN_BRACKETS, JINJA_SYMBOL_OPEN_BRACKETS)
                    + "endif"
                    + node.content[m.end(1) :].replace(GO_SYMBOL_CLOSE_BRACKETS, JINJA_SYMBOL_CLOSE_BRACKETS)
                )
            elif node.open_scope_node and node.open_scope_node.type == NodeType.RANGE:
                loop_vars.pop()
                if loop_index_vars:
                    loop_index_vars.pop()

                return (
                    node.content[: m.start(1)].replace(GO_SYMBOL_OPEN_BRACKETS, JINJA_SYMBOL_OPEN_BRACKETS)
                    + "endfor"
                    + node.content[m.end(1) :].replace(GO_SYMBOL_CLOSE_BRACKETS, JINJA_SYMBOL_CLOSE_BRACKETS)
                )
        elif node.type == NodeType.ASSIGNMENT:
            m = GO_KEYWORDS[NodeType.ASSIGNMENT].match(node.content)
            if m is not None and len(m.groups()) == 2:
                variable = m.groups()[0].strip()
                pipeline = m.groups()[1].strip()
                return (
                    node.content.replace(variable, f"set {transform_go_local_var_to_jinja(variable)}", 1)
                    .replace(pipeline, parse_pipeline(pipeline), 1)
                    .replace(":=", "=", 1)
                    .replace(GO_SYMBOL_OPEN_BRACKETS, JINJA_SYMBOL_OPEN_BRACKETS, 1)
                    .replace(GO_SYMBOL_CLOSE_BRACKETS, JINJA_SYMBOL_CLOSE_BRACKETS, 1)
                )
        elif node.type == NodeType.RANGE:
            m = GO_KEYWORDS[NodeType.RANGE].match(node.content)
            if m is not None:
                loop_var = transform_go_var_to_jinja(m.groups()[4])
                loop_vars.append(loop_var[0])
                if m.groups()[2] is not None:
                    loop_index_vars.append(m.groups()[2])
                content = (
                    node.content.replace("range", f"for {loop_var[0]} in")
                    .replace(m.groups()[4], transform_go_var_to_jinja(m.groups()[4], False))
                    .replace(GO_SYMBOL_OPEN_BRACKETS, JINJA_SYMBOL_OPEN_BRACKETS)
                    .replace(GO_SYMBOL_CLOSE_BRACKETS, JINJA_SYMBOL_CLOSE_BRACKETS)
                )
                if m.groups()[1] is not None:
                    content = content.replace(m.groups()[1], "")
                return content
        elif node.type == NodeType.CONTINUE:
            return node.content.replace(GO_SYMBOL_OPEN_BRACKETS, JINJA_SYMBOL_OPEN_BRACKETS).replace(
                GO_SYMBOL_CLOSE_BRACKETS, JINJA_SYMBOL_CLOSE_BRACKETS
            )
        elif node.type == NodeType.BREAK:
            return node.content.replace(GO_SYMBOL_OPEN_BRACKETS, JINJA_SYMBOL_OPEN_BRACKETS).replace(
                GO_SYMBOL_CLOSE_BRACKETS, JINJA_SYMBOL_CLOSE_BRACKETS
            )

        return node.content

    def nodes_to_jinja_str(nodes: list[Node]) -> str:
        res = ""
        for node in nodes:
            res += node_to_jinja_str(node)
            res += nodes_to_jinja_str(node.children)
        return res

    return nodes_to_jinja_str(parse_go_template(content))


def tree_structure(nodes: list[Node], level: int) -> str:
    res = ""
    for node in nodes:
        parent_type = "--" if node.parent is None else node.parent.type
        res += level * "\t" + f"{node.type}: {node.start},{node.end} - " f"{parent_type} - {node.content}\n"
        res += tree_structure(node.children, level + 1)

    return res


def tree_content(nodes: list[Node], level: int) -> str:
    res = ""
    for node in nodes:
        res += node.content
        res += tree_content(node.children, level + 1)
    return res


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="go2jinja",
        description="Simple Tool for converting Go Templates to Jinja Templates",
    )

    parser.add_argument(
        "--go-template",
        dest="template",
        required=True,
        help="Path to the file containing the Go Template to convert to Jinja",
    )
    parser.add_argument(
        "--output",
        dest="output",
        default="",
        help="Output file path for the converted Jinja Template. If empty, prints to stdout.",
    )

    args = parser.parse_args()

    jinja = ""
    with open(args.template, "r") as input:
        jinja = go_to_jinja(input.read())
    if args.output == "":
        print(jinja)
        sys.exit(0)
    with open(args.output, "w") as output:
        output.write(jinja)
        output.flush()



class InvalidInferenceEngineSpecError(RuntimeError):

    def __init__(self, spec_file: str, reason: str, *args):
        super().__init__(*args)

        self.spec_file = spec_file
        self.reason = reason

    def __str__(self):
        return f"Invalid spec file '{self.spec_file}': {self.reason}"



# The following code is inspired from: https://github.com/ericcurtin/lm-pull/blob/main/lm-pull.py

import os
import shutil
import sys
import time
import urllib.request

import ramalama.console as console
from ramalama.common import perror
from ramalama.config import CONFIG
from ramalama.file import File
from ramalama.logger import logger
from ramalama.proxy_support import setup_proxy_support

# Setup proxy support on module import
setup_proxy_support()

HTTP_NOT_FOUND = 404
HTTP_RANGE_NOT_SATISFIABLE = 416  # "Range Not Satisfiable" error (file already downloaded)


class HttpClient:
    def __init__(self):
        pass

    def init(self, url, headers, output_file, show_progress, response_str=None):
        output_file_partial = None
        if output_file:
            output_file_partial = output_file + ".partial"

        self.file_size = self.set_resume_point(output_file_partial)
        self.urlopen(url, headers)
        self.total_to_download = int(self.response.getheader('content-length', 0))
        if response_str is not None:
            response_str.append(self.response.read().decode('utf-8'))
        else:
            out = File()
            try:
                if not out.open(output_file_partial, "ab"):
                    raise IOError("Failed to open file")

                if out.lock():
                    raise IOError("Failed to exclusively lock file")

                self.now_downloaded = 0
                self.start_time = time.time()
                self.perform_download(out.file, show_progress)
            finally:
                del out  # Ensure file is closed before rename

        if output_file:
            os.rename(output_file_partial, output_file)

    def urlopen(self, url, headers):
        headers["Range"] = f"bytes={self.file_size}-"
        logger.debug(f"Running urlopen {url} with headers: {headers}")
        request = urllib.request.Request(url, headers=headers)
        self.response = urllib.request.urlopen(request)

        if self.response.status not in (200, 206):
            raise IOError(f"Request failed: {self.response.status}")

    def perform_download(self, file, show_progress):
        self.total_to_download += self.file_size
        self.now_downloaded = 0
        self.start_time = time.time()
        accumulated_size = 0
        last_update_time = time.time()
        try:
            while True:
                data = self.response.read(1024)
                if not data:
                    break

                size = file.write(data)
                if show_progress:
                    accumulated_size += size
                    if time.time() - last_update_time >= 0.1:
                        self.update_progress(accumulated_size)
                        accumulated_size = 0
                        last_update_time = time.time()

            if show_progress:
                if accumulated_size > 0:
                    self.update_progress(accumulated_size)
        finally:
            if show_progress:
                # Output a newline after the progress bar
                perror("")

    def human_readable_time(self, seconds):
        hrs = int(seconds) // 3600
        mins = (int(seconds) % 3600) // 60
        secs = int(seconds) % 60
        width = 10
        if hrs > 0:
            return f"{hrs}h {mins:02}m {secs:02}s".rjust(width)
        elif mins > 0:
            return f"{mins}m {secs:02}s".rjust(width)
        else:
            return f"{secs}s".rjust(width)

    def human_readable_size(self, size):
        width = 10
        for unit in ["B", "KB", "MB", "GB", "TB"]:
            if size < 1024:
                return f"{size:.2f} {unit}".rjust(width)

            size /= 1024

        return f"{size:.2f} PB".rjust(width)

    def get_terminal_width(self):
        return shutil.get_terminal_size().columns

    def generate_progress_prefix(self, percentage):
        return f"{percentage}% |".rjust(6)

    def generate_progress_suffix(self, now_downloaded_plus_file_size, speed, estimated_time):
        return f"{self.human_readable_size(now_downloaded_plus_file_size)}/{self.human_readable_size(self.total_to_download)}{self.human_readable_size(speed)}/s{self.human_readable_time(estimated_time)}"  # noqa: E501

    def calculate_progress_bar_width(self, progress_prefix, progress_suffix):
        progress_bar_width = self.get_terminal_width() - len(progress_prefix) - len(progress_suffix) - 3
        if progress_bar_width < 1:
            progress_bar_width = 1

        return progress_bar_width

    def generate_progress_bar(self, progress_bar_width, percentage):
        pos = (percentage * progress_bar_width) // 100
        progress_bar = ""
        for i in range(progress_bar_width):
            progress_bar += "█" if i < pos else " "

        return progress_bar

    def set_resume_point(self, output_file):
        if output_file and os.path.exists(output_file):
            return os.path.getsize(output_file)

        return 0

    def print_progress(self, progress_prefix, progress_bar, progress_suffix):
        perror(f"\r{progress_prefix}{progress_bar}| {progress_suffix}", end="")

    def update_progress(self, chunk_size):
        self.now_downloaded += chunk_size
        now_downloaded_plus_file_size = self.now_downloaded + self.file_size
        percentage = (now_downloaded_plus_file_size * 100) // self.total_to_download if self.total_to_download else 100
        progress_prefix = self.generate_progress_prefix(percentage)
        speed = self.calculate_speed(self.now_downloaded, self.start_time)
        tim = (self.total_to_download - self.now_downloaded) // speed
        progress_suffix = self.generate_progress_suffix(now_downloaded_plus_file_size, speed, tim)
        progress_bar_width = self.calculate_progress_bar_width(progress_prefix, progress_suffix)
        progress_bar = self.generate_progress_bar(progress_bar_width, percentage)
        self.print_progress(progress_prefix, progress_bar, progress_suffix)

    def calculate_speed(self, now_downloaded, start_time):
        now = time.time()
        elapsed_seconds = now - start_time
        return now_downloaded / elapsed_seconds


def download_file(url: str, dest_path: str, headers: dict[str, str] | None = None, show_progress: bool = True):
    """
    Downloads a file from a given URL to a specified destination path.

    Args:
        url (str): The URL to download from.
        dest_path (str): The path to save the downloaded file.
        headers (dict): Optional headers to include in the request.
        show_progress (bool): Whether to show a progress bar during download.

    Raises:
        RuntimeError: If the download fails after multiple attempts.
    """
    headers = headers or {}

    # If not running in a TTY, disable progress to prevent CI pollution
    if not sys.stdout.isatty():
        show_progress = False

    http_client = HttpClient()
    max_retries = CONFIG.http_client.max_retries
    retries = 0

    while retries <= max_retries:
        try:
            # Initialize HTTP client for the request
            http_client.init(url=url, headers=headers, output_file=dest_path, show_progress=show_progress)
            return  # Exit function if successful

        except KeyboardInterrupt:
            perror("\nDownload interrupted by user. Exiting cleanly.")
            raise

        except urllib.error.HTTPError as e:
            if e.code in [HTTP_RANGE_NOT_SATISFIABLE, HTTP_NOT_FOUND]:
                raise e
            retries += 1

        except urllib.error.URLError as e:
            console.error(f"Network Error: {e.reason}")
            retries += 1

        except TimeoutError:
            retries += 1
            console.warning(f"TimeoutError: The server took too long to respond. Retrying {retries}/{max_retries} ...")

        except RuntimeError as e:  # Catch network-related errors from HttpClient
            retries += 1
            console.warning(f"{e}. Retrying {retries}/{max_retries} ...")

        except IOError as e:
            retries += 1
            console.warning(f"I/O Error: {e}. Retrying {retries}/{max_retries} ...")

        except Exception as e:
            console.error(f"Unexpected error: {str(e)}")
            raise e

        if retries > max_retries:
            error_message = (
                "\nDownload failed after multiple attempts.\n"
                "Possible causes:\n"
                "- Internet connection issue\n"
                "- Server is down or unresponsive\n"
                "- Firewall or proxy blocking the request\n"
            )
            raise ConnectionError(error_message)

        time.sleep(
            min(CONFIG.http_client.max_retry_delay, 2 ** (retries - 1) * 0.1)
        )  # Exponential backoff (0.1s, 0.2s, 0.4s... max_retry_delay)



from dataclasses import MISSING, fields, is_dataclass
from functools import reduce
from typing import Any, Type, get_type_hints


def deep_merge(left: dict, right: dict) -> dict:
    """Recursively merge override dict into base dict."""
    for key, value in right.items():
        if isinstance(left_value := left.get(key, None), dict) and isinstance(value, dict):
            left[key] = deep_merge(left_value, value)
        else:
            left[key] = value
    return left


def extract_defaults(cls) -> dict[str, Any]:
    result = {}
    for f in fields(cls):
        if f.default is not MISSING:
            result[f.name] = f.default
        elif f.default_factory is not MISSING:  # type: ignore
            result[f.name] = f.default_factory()  # type: ignore
    return result


def build_subconfigs(values: dict, obj: Type) -> dict:
    """Facilitates nesting configs by instantiating the child typed object from a dict

    NOTE: This implementation does not automatically coerce more complicated config structures
    involving types like (ConfigObj | None), or list[ConfigObj], etc...
    """

    dtypes: dict[str, Type] = get_type_hints(obj)
    for k, v in values.items():
        if isinstance(v, dict) and (subconfig_type := dtypes.get(k)) and is_dataclass(subconfig_type):
            values[k] = subconfig_type(**build_subconfigs(v, dtypes[k]))

    return values


class LayeredMixin:
    """Mixin class to handle layered configurations from multiple sources."""

    def __init__(self, *layers: dict[str, Any]):
        self._fields = {f.name for f in fields(self.__class__)}  # type: ignore[arg-type]
        self._layers = [{k: layer[k] for k in layer.keys() & self._fields} for layer in layers]

        defaults = extract_defaults(self.__class__)
        merged = defaults | reduce(deep_merge, reversed(self._layers))
        super().__init__(**build_subconfigs(merged, type(self)))

    def is_set(self, name: str) -> bool:
        return any(name in layer for layer in self._layers)



import json
from dataclasses import dataclass

from ramalama.config import BaseConfig
from ramalama.daemon.dto.errors import MissingArgumentError


@dataclass
class ServeRequest:

    model_name: str
    runtime: str
    exec_args: dict[str, str]

    def to_dict(self) -> dict:
        return {
            "model_name": self.model_name,
            "runtime": self.runtime,
            "exec_args": dict(
                [
                    (key, value) for key, value in self.exec_args.items() if type(value) is str
                ]  # Filter out non-string values
            ),
        }

    def serialize(self) -> str:
        return json.dumps(self.to_dict(), indent=4, sort_keys=True)

    @staticmethod
    def from_string(data: str) -> "ServeRequest":
        data_dict = json.loads(data)

        model_name = data_dict.get("model_name", None)
        if not model_name:
            raise MissingArgumentError("model_name")

        runtime = data_dict.get("runtime", None)
        if not runtime:
            raise MissingArgumentError("runtime")

        base_exec_args = BaseConfig().__dict__
        exec_args_input = data_dict.get("exec_args", {})
        # merge missing args to args from base config
        exec_args = {
            **base_exec_args,
            **{
                "runtime_args": [],
                "debug": False,
            },
        }
        # overwrite with input args
        exec_args = {**exec_args, **exec_args_input}
        # overwrite certain values which do not make sense in this context
        exec_args = {
            **exec_args,
            **{
                "container": False,
                "generate": False,
                "dryrun": False,
            },
        }

        return ServeRequest(
            model_name=model_name,
            runtime=runtime,
            exec_args=exec_args,
        )


@dataclass
class ServeResponse:

    model_id: str
    serve_path: str

    def to_dict(self) -> dict:
        return {
            "model_id": self.model_id,
            "serve_path": self.serve_path,
        }

    def serialize(self) -> str:
        return json.dumps(self.to_dict(), indent=4, sort_keys=True)


@dataclass
class StopServeRequest:

    model_name: str

    @staticmethod
    def from_string(data: str) -> "StopServeRequest":
        data_dict = json.loads(data)

        model_name = data_dict.get("model_name", None)
        if not model_name:
            raise MissingArgumentError("model_name")

        return StopServeRequest(model_name=model_name)

    def to_dict(self) -> dict:
        return {
            "model_name": self.model_name,
        }

    def serialize(self) -> str:
        return json.dumps(self.to_dict(), indent=4, sort_keys=True)



import os
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, List

from ramalama import oci_tools
from ramalama.arg_types import EngineArgs
from ramalama.model_store.constants import DIRECTORY_NAME_BLOBS, DIRECTORY_NAME_REFS, DIRECTORY_NAME_SNAPSHOTS
from ramalama.model_store.reffile import RefJSONFile, migrate_reffile_to_refjsonfile


@dataclass
class ModelFile:
    name: str
    modified: float
    size: int
    is_partial: bool


class GlobalModelStore:
    def __init__(
        self,
        base_path: str,
    ):
        self._store_base_path = os.path.join(base_path, "store")

    @property
    def path(self) -> str:
        return self._store_base_path

    def list_models(self, engine: str, show_container: bool) -> Dict[str, List[ModelFile]]:
        models: Dict[str, List[ModelFile]] = {}

        for root, subdirs, _ in os.walk(self.path):
            if DIRECTORY_NAME_REFS in subdirs:
                ref_dir = os.path.join(root, DIRECTORY_NAME_REFS)
                for ref_file_name in os.listdir(ref_dir):
                    ref_file_path = os.path.join(ref_dir, ref_file_name)
                    ref_file = migrate_reffile_to_refjsonfile(
                        ref_file_path, os.path.join(root, DIRECTORY_NAME_SNAPSHOTS)
                    )
                    if ref_file is None:
                        ref_file = RefJSONFile.from_path(ref_file_path)

                    model_path = root.replace(self.path, "").replace(os.sep, "", 1)

                    parts = model_path.split("/")
                    model_source = parts[0]
                    model_path_without_source = f"{os.sep}".join(parts[1:])

                    separator = ":///" if model_source == "file" else "://"  # Use ':///' for file URLs, '://' otherwise
                    tag = ref_file_name.replace(".json", "")
                    model_name = f"{model_source}{separator}{model_path_without_source}:{tag}"

                    collected_files = []
                    for snapshot_file in ref_file.files:
                        is_partially_downloaded = False
                        snapshot_file_path = os.path.join(
                            root, DIRECTORY_NAME_SNAPSHOTS, ref_file.hash, snapshot_file.name
                        )
                        if not os.path.exists(snapshot_file_path):
                            blobs_partial_file_path = os.path.join(
                                root, DIRECTORY_NAME_BLOBS, ref_file.hash + ".partial"
                            )
                            if not os.path.exists(blobs_partial_file_path):
                                continue

                            snapshot_file_path = blobs_partial_file_path
                            is_partially_downloaded = True

                        last_modified = os.path.getmtime(snapshot_file_path)
                        file_size = os.path.getsize(snapshot_file_path)
                        collected_files.append(
                            ModelFile(snapshot_file.name, last_modified, file_size, is_partially_downloaded)
                        )
                    models[model_name] = collected_files

        if show_container:
            oci_models = oci_tools.list_models(EngineArgs(engine=engine))
            for oci_model in oci_models:
                name, modified, size = (oci_model["name"], oci_model["modified"], oci_model["size"])
                # oci_tools.list_models provides modified as timestamp string, convert it to unix timestamp
                modified_unix = datetime.fromisoformat(modified).timestamp()
                models[name] = [ModelFile(name, modified_unix, size, is_partial=False)]

        return models

    # TODO:
    # iterating over all symlinks in snapshot dir, check valid
    def verify_snapshot(self):
        pass

    # TODO:
    # iterating over models and check
    #    1. for broken symlinks in snapshot dirs -> delete and update refs
    #    2. for blobs not reached by ref->snapshot chain -> delete
    #    3. for empty folders -> delete
    def cleanup(self):
        pass



import copy
from collections.abc import Callable
from typing import TypeAlias
from urllib.parse import urlparse

from ramalama.arg_types import StoreArgType
from ramalama.common import rm_until_substring
from ramalama.config import CONFIG
from ramalama.transports.base import MODEL_TYPES
from ramalama.transports.huggingface import Huggingface
from ramalama.transports.modelscope import ModelScope
from ramalama.transports.oci import OCI
from ramalama.transports.ollama import Ollama
from ramalama.transports.rlcr import RamalamaContainerRegistry
from ramalama.transports.url import URL

CLASS_MODEL_TYPES: TypeAlias = Huggingface | Ollama | OCI | URL | ModelScope | RamalamaContainerRegistry


class TransportFactory:
    def __init__(
        self,
        model: str,
        args: StoreArgType,
        transport: str = "ollama",
        ignore_stderr: bool = False,
    ):
        self.model = model
        self.store_path = args.store
        self.transport = transport
        self.engine = args.engine
        self.ignore_stderr = ignore_stderr
        self.container = args.container

        model_cls, _create = self.detect_model_model_type()

        self.model_cls = model_cls
        self._create = _create

        self.pruned_model = self.prune_model_input()
        self.draft_model = None

        if getattr(args, 'model_draft', None):
            dm_args = copy.deepcopy(args)
            dm_args.model_draft = None  # type: ignore
            self.draft_model = TransportFactory(args.model_draft, dm_args, ignore_stderr=True).create()  # type: ignore

    def detect_model_model_type(self) -> tuple[type[CLASS_MODEL_TYPES], Callable[[], CLASS_MODEL_TYPES]]:
        for prefix in ["huggingface://", "hf://", "hf.co/"]:
            if self.model.startswith(prefix):
                return Huggingface, self.create_huggingface
        for prefix in ["modelscope://", "ms://"]:
            if self.model.startswith(prefix):
                return ModelScope, self.create_modelscope
        for prefix in ["ollama://", "ollama.com/library/"]:
            if self.model.startswith(prefix):
                return Ollama, self.create_ollama
        for prefix in ["oci://", "docker://"]:
            if self.model.startswith(prefix):
                return OCI, self.create_oci
        if self.model.startswith("rlcr://"):
            return RamalamaContainerRegistry, self.create_rlcr
        for prefix in ["http://", "https://", "file://"]:
            if self.model.startswith(prefix):
                return URL, self.create_url
        if self.transport == "huggingface":
            return Huggingface, self.create_huggingface
        if self.transport == "modelscope":
            return ModelScope, self.create_modelscope
        if self.transport == "ollama":
            return Ollama, self.create_ollama
        if self.transport == "rlcr":
            return RamalamaContainerRegistry, self.create_rlcr
        if self.transport == "oci":
            return OCI, self.create_oci

        raise KeyError(f'transport "{self.transport}" not supported. Must be oci, huggingface, modelscope, or ollama.')

    def prune_model_input(self) -> str:
        # remove protocol from model input
        pruned_model_input = rm_until_substring(self.model, "://")

        if self.model_cls == Huggingface:
            pruned_model_input = rm_until_substring(pruned_model_input, "hf.co/")
        elif self.model_cls == ModelScope:
            pruned_model_input = rm_until_substring(pruned_model_input, "modelscope.cn/")
        elif self.model_cls == Ollama:
            pruned_model_input = rm_until_substring(pruned_model_input, "ollama.com/library/")

        return pruned_model_input

    def validate_oci_model_input(self):
        if self.model.startswith("oci://") or self.model.startswith("docker://") or self.model.startswith("rlcr://"):
            return

        for t in MODEL_TYPES:
            if self.model.startswith(t + "://"):
                raise ValueError(f"{self.model} invalid: Only OCI Model types supported")

    def create(self) -> CLASS_MODEL_TYPES:
        return self._create()

    def create_huggingface(self) -> Huggingface:
        model = Huggingface(self.pruned_model, self.store_path)
        model.draft_model = self.draft_model
        return model

    def create_modelscope(self) -> ModelScope:
        model = ModelScope(self.pruned_model, self.store_path)
        model.draft_model = self.draft_model
        return model

    def create_ollama(self) -> Ollama:
        model = Ollama(self.pruned_model, self.store_path)
        model.draft_model = self.draft_model
        return model

    def create_rlcr(self) -> RamalamaContainerRegistry:
        if not self.container:
            raise ValueError("OCI containers cannot be used with the --nocontainer option.")

        if self.engine is None:
            raise ValueError("Constructing an OCI model factory requires an engine value")

        self.validate_oci_model_input()
        model = RamalamaContainerRegistry(
            model=self.pruned_model,
            model_store_path=self.store_path,
            conman=self.engine,
            ignore_stderr=self.ignore_stderr,
        )
        model.draft_model = self.draft_model
        return model

    def create_oci(self) -> OCI:
        if not self.container:
            raise ValueError("OCI containers cannot be used with the --nocontainer option.")

        if self.engine is None:
            raise ValueError("Constructing an OCI model factory requires an engine value")

        self.validate_oci_model_input()

        model = OCI(self.pruned_model, self.store_path, self.engine, self.ignore_stderr)
        model.draft_model = self.draft_model
        return model

    def create_url(self) -> URL:
        model = URL(self.pruned_model, self.store_path, urlparse(self.model).scheme)
        model.draft_model = self.draft_model
        return model


def New(name, args, transport: str | None = None) -> CLASS_MODEL_TYPES:
    if transport is None:
        transport = CONFIG.transport
    return TransportFactory(name, args, transport=transport).create()
